from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.core.errors import InvalidItemValueError, ItemValueIsNoneError
from HABApp.openhab.definitions.websockets.item_value_types import (
    PercentTypeModel,
)
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.items.commands import OnOffCommand, PercentCommand


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/DimmerItem.java
class DimmerItem(OpenhabItem, OnOffCommand, PercentCommand):
    """DimmerItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar int | float value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('DimmerItem', PercentTypeModel, 'OnOff', 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('DimmerItem', PercentTypeModel, 'OnOff', 'IncreaseDecrease', 'Refresh')
    _state_from_oh_str: Final = staticmethod(PercentTypeModel.get_value_from_state)

    def set_value(self, new_value: float | None) -> bool:

        # Percent is 0 ... 100
        if isinstance(new_value, (int, float)) and (0 <= new_value <= 100):
            return super().set_value(new_value)

        if new_value is None:
            return super().set_value(new_value)

        raise InvalidItemValueError.from_item(self, new_value)

    def is_on(self) -> bool:
        """Test value against on-value"""
        return bool(self.value)

    def is_off(self) -> bool:
        """Test value against off-value"""
        return self.value is not None and not self.value

    def __str__(self) -> str:
        return str(self.value)

    def __bool__(self) -> bool:
        if self.value is None:
            raise ItemValueIsNoneError.from_item(self)
        return self.is_on()



from __future__ import annotations

from time import monotonic


class TimeoutNotRunningError(Exception):
    pass


class Timeout:
    __slots__ = ('_timeout', '_started')

    def __init__(self, timeout: float, *, start: bool = True) -> None:
        self._timeout: float = timeout
        if self._timeout <= 0:
            raise ValueError()

        self._started: float | None = None if not start else monotonic()

    def __repr__(self) -> str:

        decimals = 1 if self._timeout < 10 else 0

        if self._started is None:
            return f'<Timeout {self._timeout:.{decimals:d}f}s>'

        time = monotonic() - self._started
        if time >= self._timeout:
            time = self._timeout
        return f'<Timeout {time:.{decimals:d}f}/{self._timeout:.{decimals:d}f}s>'

    def reset(self):
        """Reset the timeout if it is running"""
        if self._started is not None:
            self._started = monotonic()
        return self

    def start(self):
        """Start the timeout if it is not running"""
        if self._started is None:
            self._started = monotonic()
        return self

    def stop(self):
        """Stop the timeout"""
        self._started = None
        return self

    def set_timeout(self, timeout: float):
        """Set the timeout

        :param timeout: Timeout in seconds
        """
        if self._timeout <= 0:
            raise ValueError()
        self._timeout = timeout
        return self

    def is_running(self) -> bool:
        """ Return whether the timeout is running.

        :return: True if running or False
        """
        return self._started is not None

    def is_expired(self) -> bool:
        """Return whether the timeout is expired, raises an exception if the timeout is not running

        :return: True if expired else False
        """
        if self._started is None:
            raise TimeoutNotRunningError()
        return monotonic() - self._started >= self._timeout

    def is_running_and_expired(self) -> bool:
        """Return whether the timeout is running and expired

        :return: True if expired else False
        """
        return self._started is not None and monotonic() - self._started >= self._timeout

    def remaining(self) -> float:
        """Return the remaining seconds. Raises an exception if the timeout is not running

        :return: Remaining time in seconds or 0 if expired
        """
        if self._started is None:
            raise TimeoutNotRunningError()
        remaining = self._timeout - (monotonic() - self._started)
        return 0 if remaining <= 0 else remaining

    def remaining_or_none(self) -> float | None:
        """Return the remaining seconds. Raises an exception if the timeout is not running

        :return: Remaining time in seconds, 0 if expired or None if not running
        """
        if self._started is None:
            return None
        remaining = self._timeout - (monotonic() - self._started)
        return 0 if remaining <= 0 else remaining



from __future__ import annotations

from operator import gt, lt
from typing import TYPE_CHECKING, Final, TypeAlias, overload

from eascheduler.builder.helper import HINT_TIMEDELTA, get_timedelta
from typing_extensions import Self
from whenever import Instant, TimeDelta


if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import datetime as dt_datetime
    from datetime import timedelta as dt_timedelta


class InstantView:
    __slots__ = ('_instant', )

    def __init__(self, instant: Instant) -> None:
        self._instant: Final = instant

    @classmethod
    def now(cls) -> InstantView:
        """Create a new instance with the current time"""
        return cls(Instant.now())

    def add(self, *, hours: float = 0, minutes: float = 0, seconds: float = 0,
            milliseconds: float = 0, microseconds: float = 0, nanoseconds: int = 0) -> Self:
        """Add time delta

        :return: New instance
        """
        return self.__class__(self._instant + TimeDelta(
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            milliseconds=milliseconds,
            microseconds=microseconds,
            nanoseconds=nanoseconds,
        ))

    def subtract(self, *, hours: float = 0, minutes: float = 0, seconds: float = 0,
                 milliseconds: float = 0, microseconds: float = 0, nanoseconds: int = 0) -> Self:
        """Subtract time delta from current instant

        :return: New instance
        """
        return self.__class__(self._instant + TimeDelta(
            hours=-hours,
            minutes=-minutes,
            seconds=-seconds,
            milliseconds=-milliseconds,
            microseconds=-microseconds,
            nanoseconds=-nanoseconds,
        ))

    def delta_now(self, now: InstantView | Instant | None = None) -> TimeDelta:
        """Return the delta between now and the instant.
        The delta will be positive, e.g.
        if the InstantView is from 5 seconds ago this will return a timedelta with 5 seconds.

        :param now: optional instant to compare to instead of now, must be newer than the instant of the instant view
        """

        match now:
            case None:
                ref = Instant.now()
            case InstantView():
                ref = now._instant
            case Instant():
                ref = now
            case _:
                msg = f'Invalid type: {type(now).__name__}'
                raise TypeError(msg)

        if ref < self._instant:
            msg = f'Reference instant must be newer than the instant of the {self.__class__.__name__}'
            raise ValueError(msg)

        return ref - self._instant

    def py_timedelta(self, now: InstantView | Instant | None = None) -> dt_timedelta:
        """Return the timedelta between the instant and now

        :param now: optional instant to compare to
        """
        return self.delta_now(now).py_timedelta()

    def py_datetime(self) -> dt_datetime:
        """Return the datetime of the instant"""
        return self._instant.to_system_tz().to_plain().py_datetime()

    def __repr__(self) -> str:
        return f'InstantView({self._instant.to_system_tz()})'

    def _cmp(self, op: Callable[[Instant, Instant], bool], obj: HINT_OBJ | None, *,
             delta_kwargs: dict[str, float] | None = None) -> bool:

        match obj:
            case None:
                if days := delta_kwargs.pop('days', 0):
                    delta_kwargs['hours'] = delta_kwargs.get('hours', 0) + days * 24
                td = TimeDelta(**delta_kwargs)
            case InstantView():
                return op(self._instant, obj._instant)
            case Instant():
                return op(self._instant, obj)
            case _:
                td = get_timedelta(obj)

        if td <= TimeDelta.ZERO:
            msg = 'Delta must be positive since instant is in the past'
            raise ValueError(msg)

        return op(self._instant, Instant.now() - td)

    @overload
    def older_than(self, *, days: float = 0, hours: float = 0, minutes: float = 0, seconds: float = 0,
                   milliseconds: float = 0, microseconds: float = 0, nanoseconds: int = 0) -> bool: ...

    @overload
    def older_than(self, obj: HINT_OBJ) -> bool: ...

    def older_than(self, obj=None, **kwargs):
        """Check if the instant is older than the given value"""
        return self._cmp(lt, obj, delta_kwargs=kwargs)

    @overload
    def newer_than(self, *, days: float = 0, hours: float = 0, minutes: float = 0, seconds: float = 0,
                   milliseconds: float = 0, microseconds: float = 0, nanoseconds: int = 0) -> bool: ...

    @overload
    def newer_than(self, obj: HINT_OBJ) -> bool: ...

    def newer_than(self, obj=None, **kwargs):
        """Check if the instant is newer than the given value"""
        return self._cmp(gt, obj, delta_kwargs=kwargs)

    # ------------------------------------------------------------------------------------------------------------------
    # Comparisons
    def __lt__(self, other: InstantView | Instant) -> bool:
        if isinstance(other, InstantView):
            return self._instant < other._instant
        if isinstance(other, Instant):
            return self._instant < other
        return NotImplemented

    def __le__(self, other: InstantView | Instant) -> bool:
        if isinstance(other, InstantView):
            return self._instant <= other._instant
        if isinstance(other, Instant):
            return self._instant <= other
        return NotImplemented

    def __gt__(self, other: InstantView | Instant) -> bool:
        if isinstance(other, InstantView):
            return self._instant > other._instant
        if isinstance(other, Instant):
            return self._instant > other
        return NotImplemented

    def __ge__(self, other: InstantView | Instant) -> bool:
        if isinstance(other, InstantView):
            return self._instant >= other._instant
        if isinstance(other, Instant):
            return self._instant >= other
        return NotImplemented

    def __eq__(self, other: InstantView | Instant) -> bool:
        if isinstance(other, InstantView):
            return self._instant == other._instant
        if isinstance(other, Instant):
            return self._instant == other
        return NotImplemented

    # ------------------------------------------------------------------------------------------------------------------
    # Operations
    def __add__(self, other: HINT_TIMEDELTA) -> Self:
        """Add time delta to current instant

        :param other: time delta object
        :return: New instance
        """
        try:
            td = get_timedelta(other)
        except TypeError:
            return NotImplemented
        return self.__class__(self._instant + td)

    def __sub__(self, other: HINT_TIMEDELTA) -> Self:
        """Subtract time delta from current instant

        :param other: time delta object
        :return: New instance
        """
        try:
            td = get_timedelta(other)
        except TypeError:
            return NotImplemented
        return self.__class__(self._instant - td)


HINT_OBJ: TypeAlias = HINT_TIMEDELTA | Instant | InstantView



from __future__ import annotations

import logging
from collections.abc import Callable

from typing_extensions import override

from HABApp.core.asyncio import create_task
from HABApp.core.internals import Context
from HABApp.core.internals.wrapped_function.base import P, R, WrappedFunctionBase


class WrappedSyncFunction(WrappedFunctionBase[P, R]):

    def __init__(self, func: Callable,
                 warn_too_long=True,
                 name: str | None = None,
                 logger: logging.Logger | None = None,
                 context: Context | None = None) -> None:

        super().__init__(name=name, func=func, logger=logger, context=context)
        assert callable(func)

        self.func = func
        self.warn_too_long: bool = warn_too_long

    @override
    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
        create_task(self.async_run(*args, **kwargs), name=self.name)

    @override
    async def async_run(self, *args: P.args, **kwargs: P.kwargs) -> R | None:

        try:
            return self.func(*args, **kwargs)
        except Exception as e:
            self.process_exception(e, *args, **kwargs)
            return None



from __future__ import annotations

import asyncio
import logging
from typing import Final

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.config.models.openhab import General as OpenHABGeneralConfig
from HABApp.core import shutdown
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.lib import Timeout, ValueChange
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.connection.handler.func_async import async_get_system_info


class WaitForStartlevelPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None) -> None:
        super().__init__(name)

    async def on_connected(self, context: OpenhabContext, connection: OpenhabConnection) -> None:
        if not context.is_oh41:
            return await self.__on_connected_old(context, connection)

        return await self.__on_connected_new(context, connection)

    async def __on_connected_new(self, context: OpenhabContext, connection: OpenhabConnection) -> None:
        log = connection.log
        oh_general = HABApp.CONFIG.openhab.general
        start_level_reached_timeout: Final = 10 * 60

        if (system_info := await async_get_system_info()) is not None:
            # If openHAB is already running we have a fast exit path here

            uptime_skip_always = 2 * (oh_general.min_uptime + start_level_reached_timeout)
            started_long_ago = system_info.uptime > uptime_skip_always
            if started_long_ago:
                log.debug(f'Uptime greater than {uptime_skip_always:d} secs, skipping start level check')

            # If openHAB is already running we have a fast exit path here
            if started_long_ago or (system_info.uptime >= oh_general.min_uptime and
                                    system_info.start_level >= oh_general.min_start_level):

                # Show a hint in case it's possible to increase the start level
                # A higher start level means a more consistent startup and thus is more desirable
                if system_info.start_level > oh_general.min_start_level:
                    _field_name_cfg = 'min_start_level'
                    if (alias := OpenHABGeneralConfig.model_fields[_field_name_cfg].alias) is not None:
                        _field_name_cfg = alias

                    logging.getLogger('HABApp').info(
                        f'Openhab reached start level {system_info.start_level:d} but HABApp only waits until '
                        f'level {oh_general.min_start_level:d} is reached. '
                        f'Consider increasing "{_field_name_cfg:s}" in the HABApp configuration. '
                    )

                context.waited_for_openhab = False
                return None

        log.info('Waiting for openHAB startup to be complete')
        context.waited_for_openhab = True

        timeout_start_at_level = 70
        timeout = Timeout(start_level_reached_timeout, start=False)

        level_change: ValueChange[int] = ValueChange()

        sleep_secs = 1

        while not shutdown.is_requested():
            await asyncio.sleep(sleep_secs)
            sleep_secs = 1

            if (system_info := await async_get_system_info()) is None:
                if level_change.set_missing().changed:
                    log.debug('Start level: not received!')
                continue

            level = system_info.start_level

            # Wait for min uptime
            if system_info.uptime < (min_uptime := oh_general.min_uptime):
                sleep_secs = min_uptime - system_info.uptime
                log.debug(f'Waiting {sleep_secs:d} secs until openHAB uptime of {min_uptime:d} secs is reached')
                continue

            # timeout is running
            if timeout.is_running_and_expired():
                log.warning(f'Starting even though openHAB is not ready yet (start level: {level})')
                break

            # log only when level changed, so we don't spam the log
            if level_change.set_value(level).changed:

                log.debug(f'Start level: {level:d}')
                if level >= oh_general.min_start_level:
                    break

                # Wait but start eventually because sometimes we have a bad configured thing or an offline gateway
                # that prevents the start level from advancing
                # This is a safety net, so we properly start e.g. after a power outage
                # When starting manually one should fix the blocking thing
                if level >= timeout_start_at_level:
                    timeout.start()
                    log.debug('Starting start level timeout')

        if shutdown.is_requested():
            return None
        log.info('openHAB startup complete')

    async def __on_connected_old(self, context: OpenhabContext, connection: OpenhabConnection):

        level_reached, level = await _start_level_reached()

        if level_reached:
            context.waited_for_openhab = False
            return None

        context.waited_for_openhab = False
        log = connection.log

        log.info('Waiting for openHAB startup to be complete')

        last_level: int = -100

        timeout_start_at_level = 70
        timeout = Timeout(10 * 60, start=False)

        while not level_reached:
            await asyncio.sleep(1)

            level_reached, level = await _start_level_reached()

            # show start level change
            if last_level != level:
                if level is None:
                    log.debug('Start level: not received!')
                    level = -10
                else:
                    log.debug(f'Start level: {level}')

                # Wait but start eventually because sometimes we have a bad configured thing or an offline gateway
                # that prevents the start level from advancing
                # This is a safety net, so we properly start e.g. after a power outage
                # When starting manually one should fix the blocking thing
                if level >= timeout_start_at_level:
                    timeout.start()
                    log.debug('Starting start level timeout')

            # timeout is running
            if timeout.is_running_and_expired():
                log.warning(f'Starting even though openHAB is not ready yet (start level: {level})')
                break

            # update last level!
            last_level = level

        log.info('openHAB startup complete')


async def _start_level_reached() -> tuple[bool, None | int]:
    start_level_min = HABApp.CONFIG.openhab.general.min_start_level

    if (system_info := await async_get_system_info()) is None:
        return False, None

    start_level_is = system_info.start_level

    return start_level_is >= start_level_min, start_level_is



import asyncio
import sys
import threading
from pathlib import Path
from typing import Final

import colorama
import pydantic


def __get_library_paths() -> tuple[str, ...]:
    found: set[str] = set()
    exec_folter = Path(sys.executable).parent
    found.add(str(exec_folter))

    # detect virtual environment
    if exec_folter.with_name('pyvenv.cfg').is_file():
        folder_names = ('bin', 'include', 'lib', 'lib64', 'scripts')
        for p in exec_folter.parent.iterdir():
            if p.name.lower() in folder_names and p.is_dir():
                found.add(str(p))

    # threading and asyncio are sometimes in a system-wide directory,
    # Pydantic and colorama should always be in the venv
    for m in (threading, asyncio, pydantic, colorama):
        f = Path(m.__file__)
        # sometimes it's a package, sometimes a module
        if f.name == '__init__.py':
            f = f.parent
        found.add(str(f.parent))

    # check if we already have the parent path
    ret = []
    for p in sorted(found):
        for parent in found:
            if p != parent and p.startswith(parent):
                break
        else:
            ret.append(p)

    return tuple(ret)


def _get_habapp_module_path() -> str:
    this = Path(__file__)
    while this.name != 'HABApp' or not this.is_dir():
        this = this.parent
    return str(this)


PYTHON_INSTALLATION_PATHS: Final = __get_library_paths()
HABAPP_MODULE_PATH: Final = _get_habapp_module_path()

del __get_library_paths
del _get_habapp_module_path



import faulthandler
import logging
import signal
from asyncio import sleep
from datetime import datetime
from typing import Final, TextIO

import HABApp
from HABApp.config.logging import rotate_file
from HABApp.core.asyncio import create_task
from HABApp.core.wrapper import log_exception


TRACEBACK_FILE: TextIO | None = None


def setup_debug() -> None:
    global TRACEBACK_FILE

    debug = HABApp.CONFIG.habapp.debug
    periodic_tb_cfg = debug.periodic_traceback
    event_loop_cfg = debug.watch_event_loop
    tb_on_shutdown_signal = debug.traceback_on_shutdown_signal

    if not event_loop_cfg.enabled and not tb_on_shutdown_signal and not periodic_tb_cfg.enabled:
        return None

    file: Final = HABApp.CONFIG.directories.logging / 'HABApp_traceback.log'
    logging.getLogger('HABApp').info(f'Dumping traceback to {file}')

    rotate_file(file, 3)

    TRACEBACK_FILE = file.open('a')
    HABApp.core.shutdown.register(TRACEBACK_FILE.close, last=True, msg='Closing traceback file')

    if periodic_tb_cfg.enabled:
        task = create_task(
            dump_traceback_task(
                int(periodic_tb_cfg.delay.total_seconds()),
                int(periodic_tb_cfg.interval.total_seconds())
            ),
            name='DumpTracebackTask'
        )
        HABApp.core.shutdown.register(task.cancel, msg='Stopping traceback task')

    if tb_on_shutdown_signal:
        TRACEBACK_FILE.write('Dumping on shutdown signal\n')
        TRACEBACK_FILE.flush()

        faulthandler.register(signal.SIGINT, TRACEBACK_FILE, all_threads=True)
        faulthandler.register(signal.SIGTERM, TRACEBACK_FILE, all_threads=True)

    if event_loop_cfg.enabled:
        task = create_task(
            watch_event_loop_task(
                sleep_secs=int(event_loop_cfg.reset_every.total_seconds()),
                timeout_secs=int(event_loop_cfg.timeout.total_seconds()),
            ),
            name='WatchEventLoopTask'
        )
        HABApp.core.shutdown.register(task.cancel, msg='Stopping WatchEventLoopTask')


@log_exception
async def dump_traceback_task(delay: int, interval: int) -> None:

    await sleep(0.2)
    TRACEBACK_FILE.write('Dumping traceback\n')
    TRACEBACK_FILE.write(f'Start: {datetime.now()}\n')
    TRACEBACK_FILE.write(f'Delay: {delay:d}s Interval: {interval:d}s\n')
    TRACEBACK_FILE.write(f'\n{"-" * 80}\n')
    TRACEBACK_FILE.flush()

    await sleep(delay)

    while True:
        TRACEBACK_FILE.write(f'{datetime.now()}\n\n')
        faulthandler.dump_traceback(TRACEBACK_FILE, all_threads=True)
        TRACEBACK_FILE.write(f'\n{"-" * 80}\n')

        await sleep(interval)


@log_exception
async def watch_event_loop_task(sleep_secs: int, timeout_secs: int) -> None:

    TRACEBACK_FILE.write(f'Watching event loop\nReset: {sleep_secs:d}s Timeout: {timeout_secs:d}s\n\n')
    TRACEBACK_FILE.flush()

    while True:
        faulthandler.dump_traceback_later(timeout_secs, file=TRACEBACK_FILE, exit=True)
        await sleep(sleep_secs)



from typing_extensions import Self, override

from HABApp.core.errors import (
    InvalidItemValueError,
    ItemNameNotOfTypeStrError,
    ItemNotFoundException,
    ItemValueIsNoneError,
    WrongItemTypeError,
)
from HABApp.core.internals import uses_item_registry
from HABApp.core.items import BaseValueItem
from HABApp.core.types import HSB, RGB


item_registry = uses_item_registry()


class ColorItem(BaseValueItem):
    """Item for dealing with color related values"""

    value: HSB | None

    @property
    def hsb(self) -> HSB:
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v

    @property
    def hue(self) -> float:
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.hue

    @property
    def saturation(self) -> float:
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.saturation

    @property
    def brightness(self) -> float:
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.brightness

    @override
    def set_value(self, new_value: RGB | HSB | tuple[float, float, float]) -> bool:
        """Set a new color value without creating events on the event bus

        :param new_value: new value of the item
        :return: True if state has changed
        """

        if isinstance(new_value, HSB):
            hsb = new_value
        elif isinstance(new_value, RGB):
            hsb = new_value.to_hsb()
        elif isinstance(new_value, tuple):
            # Convert tuple to HSB
            hue, saturation, brightness = new_value
            hsb = HSB(hue, saturation, brightness)
        elif new_value is None:
            hsb = None
        else:
            raise InvalidItemValueError.from_item(self, new_value)

        return super().set_value(new_value=hsb)

    def get_rgb(self) -> RGB:
        """Return a rgb equivalent of the color

        :return: rgb tuple
        """
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.to_rgb()

    def is_on(self) -> bool:
        """Return true if item is on"""
        if self.value is None:
            return False
        return self.value.brightness > 0

    def is_off(self) -> bool:
        """Return true if item is off"""
        if self.value is None:
            return False
        return self.value.brightness <= 0

    def __repr__(self) -> str:
        if self.value is None:
            return '<Color None>'
        return f'<Color hue: {self.hue}°, saturation: {self.saturation}%, brightness: {self.brightness}%>'

    @classmethod
    def get_create_item(cls, name: str,
                        initial_value: RGB | HSB | tuple[float, float, float] | None = None) -> Self:
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param initial_value: state the item will have if it gets created
        :param hue: hue (if initial_value is not used)
        :param saturation: saturation (if initial_value is not used)
        :param brightness: brightness (if initial_value is not used)
        :return: item
        """
        if not isinstance(name, str):
            raise ItemNameNotOfTypeStrError.from_value(name)

        try:
            item = item_registry.get_item(name)
        except ItemNotFoundException:
            item = item_registry.add_item(cls(name, initial_value))

        if not isinstance(item, cls):
            raise WrongItemTypeError.from_item(item, cls)
        return item



from typing import Any as _Any

from typing_extensions import Self as _Self

from HABApp.core.const.hints import HasNameAttr as _HasNameAttr


class HABAppException(Exception):
    pass


class ProxyObjHasNotBeenReplacedError(HABAppException):
    def __init__(self, obj) -> None:
        super().__init__(f'{obj} has not been replaced on startup!')


class ItemNotFoundException(HABAppException):
    def __init__(self, name: str) -> None:
        super().__init__(f'Item {name} does not exist!')
        self.name: str = name


class ItemAlreadyExistsError(HABAppException):
    def __init__(self, name: str) -> None:
        super().__init__(f'Item {name} does already exist and can not be added again!')
        self.name: str = name


class ContextNotFoundError(HABAppException):
    pass


class ContextNotSetError(HABAppException):
    pass


class ContextBoundObjectIsAlreadyLinkedError(HABAppException):
    pass


class ContextBoundObjectIsAlreadyUnlinkedError(HABAppException):
    pass


# ----------------------------------------------------------------------------------------------------------------------
# Value errors
# ----------------------------------------------------------------------------------------------------------------------
class HABAppValueError(ValueError, HABAppException):
    pass


class ItemValueIsNoneError(HABAppValueError):
    @classmethod
    def from_item(cls, item: _HasNameAttr) -> _Self:
        return cls(f'Item value is None (item "{item.name:s}")')


class InvalidItemValueError(HABAppValueError):
    @classmethod
    def from_item(cls, item: _HasNameAttr, value) -> _Self:
        return cls(f'Invalid value for {item.__class__.__name__} {item.name:s}: {value}')


# ----------------------------------------------------------------------------------------------------------------------
# Type errors
# ----------------------------------------------------------------------------------------------------------------------
class HABAppTypeError(TypeError, HABAppException):
    pass


class ItemNameNotOfTypeStrError(HABAppTypeError):
    @classmethod
    def from_value(cls, value: _Any) -> _Self:
        return cls(f'Item name must be of type str, not {type(value)}')


class WrongItemTypeError(HABAppTypeError):
    @classmethod
    def from_item(cls, item: _HasNameAttr, expected: type[object]) -> _Self:
        return cls(f'Item must be {expected.__name__}, not {item.__class__.__name__}')



from collections.abc import Generator
from typing import Final, Generic, Literal, TypeVar, overload

from typing_extensions import Self
from whenever import Instant, TimeDelta

from HABApp.core.const import MISSING, MISSING_TYPE


K = TypeVar('K')
V = TypeVar('V')


class ExpiringCache(Generic[K, V]):
    def __init__(self, expiry_time: float) -> None:
        self._cache: Final[dict[K, tuple[V, Instant]]] = {}
        self._expiry_time: TimeDelta = TimeDelta.ZERO

        self.set_expiry_time(expiry_time)

    def set_expiry_time(self, expiry_time: float | TimeDelta) -> None:
        """Set the expiry time for cache entries.

        :param expiry_time: expire duration
        """

        if isinstance(expiry_time, (int, float)):
            expiry_time = TimeDelta(seconds=expiry_time)

        if not isinstance(expiry_time, TimeDelta):
            raise TypeError()

        if expiry_time <= TimeDelta.ZERO:
            msg = 'Expiry time must be greater than 0'
            raise ValueError(msg)

        self._expiry_time = expiry_time

    def flush(self) -> Self:
        """Flush all expired entries out of the cache"""

        rem = []
        min_age = Instant.now() - self._expiry_time
        for key, (_, instant) in self._cache.items():
            if instant < min_age:
                rem.append(key)

        for key in rem:
            del self._cache[key]

        return self

    def reset(self, key: K) -> Self:
        """Reset the expiry time of a cache entry (if available)

        :param key: key of entry
        """
        if (obj := self._cache.get(key)) is None:
            return self

        self._cache[key] = obj[0], Instant.now()
        return self

    def set_expired(self, key: K) -> Self:
        """Set a cache entry expired (if available)

        :param key: key of entry
        """
        if (obj := self._cache.get(key)) is None:
            return self

        self._cache[key] = obj[0], Instant.now().subtract(nanoseconds=1) - self._expiry_time
        return self

    def is_expired(self, key: K) -> bool:
        """Check if a cache entry is expired

        :param key: key of entry
        """
        value, instant = self._cache[key]
        return Instant.now() > instant + self._expiry_time

    def in_cache(self, key: K) -> bool:
        """Check if a cache entry is in the cache

        :param key: key of entry
        """
        return key in self._cache

    def __iter(self, mode: Literal['all', 'expired', 'not_expired']) -> Generator[tuple[K, V], None, None]:
        if mode == 'all':
            for key, (value, _) in self._cache.items():
                yield key, value
            return None

        ret_expired = mode == 'expired'

        for key, (value, instant) in self._cache.items():
            is_expired = instant < Instant.now() - self._expiry_time

            if ret_expired:
                is_expired = not is_expired

            if not is_expired:
                yield key, value

    def set(self, key: K, value: V) -> Self:
        """Set a value in the cache

        :param key: key
        :param value: value
        """
        self._cache[key] = (value, Instant.now())
        return self

    def __setitem__(self, key: K, value: V) -> None:
        self._cache[key] = (value, Instant.now())

    @overload
    def get(self, key: K, default: V) -> V:
        ...

    @overload
    def get(self, key: K, default: None = None) -> V | None:
        ...

    def get(self, key: K, default: V | None = None) -> V | None:
        """Get a value from the cache, or return the default value if not found or expired

        :param key: key
        :param default: default
        """
        if (obj := self._cache.get(key)) is None:
            return default

        value, instant = obj
        if Instant.now() <= instant + self._expiry_time:
            return value

        return default

    @overload
    def pop(self, key: K) -> V:
        ...

    @overload
    def pop(self, key: K, default: V) -> V:
        ...

    @overload
    def pop(self, key: K, default: None) -> V | None:
        ...

    def pop(self, key: K, default: V | None | MISSING_TYPE = MISSING) -> V | None:
        """Get a value from the cache, or return the default value if not found or expired

        :param key: key
        :param default: optional default
        :return:
        """
        if (obj := self._cache.pop(key, MISSING)) is not MISSING:
            value, instant = obj
            if Instant.now() <= instant + self._expiry_time:
                return value

        if default is MISSING:
            raise KeyError(key)
        return default

    # ------------------------------------------------------------------------------------------------------------------
    # Dict interface only for values that are not expired
    # ------------------------------------------------------------------------------------------------------------------
    def __getitem__(self, key: K) -> V:
        """Get the value from the cache."""
        value, instant = self._cache[key]
        if Instant.now() <= instant + self._expiry_time:
            return value

        raise KeyError(key)

    def __contains__(self, key: K) -> bool:
        if (obj := self._cache.get(key)) is None:
            return False
        return Instant.now() <= obj[1] + self._expiry_time

    # ------------------------------------------------------------------------------------------------------------------
    # Dict interface
    # ------------------------------------------------------------------------------------------------------------------
    def __len__(self) -> int:
        return len(self._cache)

    def __delitem__(self, key: K) -> None:
        del self._cache[key]

    def __iter__(self) -> Generator[K, None, None]:
        for k, _ in self.__iter('not_expired'):
            yield k

    def keys(self, mode: Literal['all', 'expired', 'not_expired'] = 'not_expired') -> Generator[K, None, None]:
        """Get all keys in the cache that are not expired

        :param mode: ``not_expired`` - only keys of items which are not expired,
                     ``expired`` - only expired items,
                     ``all`` - all items
        """
        for k, _ in self.__iter(mode):
            yield k

    def values(self, mode: Literal['all', 'expired', 'not_expired'] = 'not_expired') -> Generator[V, None, None]:
        """Get all values in the cache that are not expired

        :param mode: ``not_expired`` - only values of items which are not expired,
                     ``expired`` - only expired items,
                     ``all`` - all items
        """
        for _, v in self.__iter(mode):
            yield v

    def items(self,
              mode: Literal['all', 'expired', 'not_expired'] = 'not_expired') -> Generator[tuple[K, V], None, None]:
        """Get all items in the cache that are not expired

        :param mode: ``not_expired`` - only items which are not expired,
                     ``expired`` - only expired items,
                     ``all`` - all items
        """
        yield from self.__iter(mode)



from __future__ import annotations

from typing_extensions import Self


class StringList(tuple[str, ...]):
    """A frozen list where all entries are of type str"""

    __slots__ = ()

    def __new__(cls, iterable: tuple[str, ...] | list[str]) -> Self:
        for v in iterable:
            if not isinstance(v, str):
                msg = f'Expected str, got {type(v)} for {cls.__name__}'
                raise TypeError(msg)
        return super().__new__(cls, iterable)

    def __eq__(self, other: tuple[str, ...] | list[str]) -> bool:
        if isinstance(other, tuple):
            return super().__eq__(other)
        if isinstance(other, list):
            return super().__eq__(tuple(other))
        return NotImplemented

    def __ne__(self, other: tuple[str, ...] | list[str]) -> bool:
        if isinstance(other, tuple):
            return super().__ne__(other)
        if isinstance(other, list):
            return super().__ne__(tuple(other))
        return NotImplemented



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.core.errors import InvalidItemValueError
from HABApp.openhab.definitions.websockets.item_value_types import (
    PercentTypeModel,
    RefreshTypeModel,
    StopMoveTypeModel,
    UnDefTypeModel,
    UpDownTypeModel,
)
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.items.commands import PercentCommand, UpDownCommand


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/RollershutterItem.java
class RollershutterItem(OpenhabItem, UpDownCommand, PercentCommand):
    """RollershutterItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar int | float value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('RollershutterItem', PercentTypeModel, UpDownTypeModel, UnDefTypeModel)
    _command_to_oh: Final = OutgoingCommandEvent(
        'RollershutterItem', UpDownTypeModel, StopMoveTypeModel, PercentTypeModel, RefreshTypeModel)
    _state_from_oh_str: Final = staticmethod(PercentTypeModel.get_value_from_state)

    def set_value(self, new_value) -> bool:

        # Position is 0 ... 100
        if isinstance(new_value, (int, float)) and (0 <= new_value <= 100):
            return super().set_value(new_value)

        if new_value is None:
            return super().set_value(new_value)

        raise InvalidItemValueError.from_item(self, new_value)

    def is_up(self) -> bool:
        return self.value <= 0

    def is_down(self) -> bool:
        return self.value >= 100

    def __str__(self) -> str:
        return str(self.value)



import logging
from typing import Any

from immutables import Map

from HABApp.core.wrapper import process_exception
from HABApp.openhab.items import (
    CallItem,
    ColorItem,
    ContactItem,
    DatetimeItem,
    DimmerItem,
    GroupItem,
    ImageItem,
    LocationItem,
    NumberItem,
    PlayerItem,
    RollershutterItem,
    StringItem,
    SwitchItem,
)
from HABApp.openhab.items.base_item import HINT_TYPE_OPENHAB_ITEM, MetaData, OpenhabItem


log = logging.getLogger('HABApp.openhab.items')


_items: dict[str, HINT_TYPE_OPENHAB_ITEM] = {
    'String': StringItem,
    'Number': NumberItem,
    'Switch': SwitchItem,
    'Contact': ContactItem,
    'Rollershutter': RollershutterItem,
    'Dimmer': DimmerItem,
    'DateTime': DatetimeItem,
    'Color': ColorItem,
    'Image': ImageItem,
    'Group': GroupItem,
    'Player': PlayerItem,
    'Location': LocationItem,
    'Call': CallItem,
}


def map_item(name: str, type: str, value: str | None,
             label: str | None, tags: frozenset[str],
             groups: frozenset[str], metadata: dict[str, dict[str, Any]] | None) -> \
        OpenhabItem | None:
    try:
        assert isinstance(type, str)
        assert value is None or isinstance(value, str)

        kwargs = {}

        # map Metadata
        if metadata is not None:
            meta = Map({k: MetaData(v['value'], Map(v.get('config', {}))) for k, v in metadata.items()})
        else:
            meta = Map()

        # Quantity types are like this: Number:Temperature and have a unit set: "12.3 °C".
        # We have to remove the dimension from the type and remove the unit from the value
        if type.startswith('Number:'):
            type, dimension = type.split(':')
            kwargs['dimension'] = dimension

            # Show warning
            # https://github.com/spacemanspiff2007/HABApp/issues/383
            if 'unit' not in meta:
                log.warning(f'Item {name:s} is a UoM item but "unit" is not found in item metadata')

        cls = _items.get(type)
        if cls is not None:
            return cls.from_oh(name, value, label=label, tags=tags, groups=groups, metadata=meta, **kwargs)

        msg = f'Unknown openHAB type: {type} for {name}'
        raise ValueError(msg)  # noqa: TRY301

    except Exception as e:
        process_exception(map_item, e, logger=log)
        return None



from collections.abc import Generator, Iterable
from pathlib import Path
from re import Pattern
from typing import Final


class FileNameBuilderRule:
    def __init__(self, prefix: str, folder: str, *,
                 priority: int, pattern: Pattern | None = None) -> None:
        self.prefix: Final = prefix
        self.folder: Final = folder
        self.priority: Final = priority
        self.pattern: Final = pattern

    def create_name(self, path: str) -> str | None:
        if not path.startswith(folder := self.folder):
            return None

        if (p := self.pattern) is not None and not p.search(path):
            return None
        return self.prefix + path.removeprefix(folder)

    def create_path(self, name: str) -> Path | None:
        if not name.startswith(prefix := self.prefix):
            return None

        if (p := self.pattern) is not None and not p.search(name):
            return None

        return Path(self.folder + name.removeprefix(prefix))

    def matches_name(self, name: str) -> bool:
        return name.startswith(self.prefix) and (self.pattern is None or self.pattern.search(name))


class FileNameBuilder:
    def __init__(self) -> None:
        self._builders: tuple[FileNameBuilderRule, ...] = ()

    def add_folder(self, prefix: str, folder: Path, *,
                   priority: int, pattern: Pattern | None = None) -> None:
        for b in self._builders:
            if b.priority == priority:
                msg = f'Priority {priority} already exists for {b.prefix}!'
                raise ValueError(msg)

        new = FileNameBuilderRule(prefix, folder.as_posix() + '/', priority=priority, pattern=pattern)
        self._builders = tuple(sorted(self._builders + (new,), key=lambda x: x.priority, reverse=True))

    def create_name(self, path: str) -> str:
        paths = [n for b in self._builders if (n := b.create_name(path)) is not None]
        if not paths:
            msg = f'Nothing matched for path {path:s}'
            raise ValueError(msg)

        if len(paths) > 1:
            msg = f'Multiple matches for path {path:s}: {", ".join(paths)}'
            raise ValueError(msg)

        return paths[0]

    def create_path(self, name: str) -> Path:
        paths = [p for b in self._builders if (p := b.create_path(name)) is not None]
        if not paths:
            msg = f'Nothing matched for name {name:s}'
            raise ValueError(msg)

        if len(paths) > 1:
            msg = f'Multiple matches for name {name:s}: {", ".join(p.as_posix() for p in paths)}'
            raise ValueError(msg)

        return paths[0]

    def is_accepted_path(self, path: str) -> bool:
        return any(b.create_name(path) is not None for b in self._builders)

    def is_accepted_name(self, path: str) -> bool:
        return any(b.matches_name(path) for b in self._builders)

    def get_folders(self) -> list[str]:
        ret: list[str] = []
        for b in self._builders:
            if b.folder not in ret:
                ret.append(b.folder)
        return ret

    def get_names_with_path(self, paths: list[str]) -> list[tuple[str, Path]]:
        ret = []
        for b in self._builders:
            ret.extend((n, Path(p)) for p in paths if (n := b.create_name(p)) is not None)
        return ret

    def get_names(self, names: Iterable[str], *, reverse: bool = False) -> Generator[str, None, None]:
        """Get sorted names"""

        names = sorted(names, reverse=reverse)
        for b in self._builders if not reverse else reversed(self._builders):
            for name in names:
                if b.matches_name(name):
                    yield name



from HABApp.config.logging.handler import CompressedMidnightRotatingFileHandler, MidnightRotatingFileHandler
from HABApp.config.logging.utils import rotate_file


# isort: split

from HABApp.config.logging.config import get_logging_dict, inject_queue_handler, load_logging_file
from HABApp.config.logging.default_logfile import create_default_logfile, get_default_logfile
from HABApp.config.logging.queue_handler import HABAppQueueHandler



from __future__ import annotations

import logging.handlers
import signal
import traceback
from asyncio import iscoroutinefunction, sleep
from dataclasses import dataclass
from types import BuiltinMethodType, FunctionType, MethodType
from typing import TYPE_CHECKING

from HABApp.core.asyncio import create_task
from HABApp.core.const import loop
from HABApp.core.lib.helper import get_obj_name


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Any, NoReturn


log = logging.getLogger('HABApp.Shutdown')


@dataclass(frozen=True)
class ShutdownBase:
    msg: str
    last: bool

    async def run(self) -> NoReturn:
        raise NotImplementedError()


@dataclass(frozen=True)
class ShutdownFunction(ShutdownBase):
    func: Callable[[], Any]

    async def run(self) -> None:
        self.func()


@dataclass(frozen=True)
class ShutdownAwaitable(ShutdownBase):
    func: Callable[[], Awaitable[Any]]

    async def run(self) -> None:
        await self.func()


_REGISTERED: tuple[ShutdownFunction | ShutdownAwaitable, ...] = ()

_REQUESTED: bool = False


def register(func: Callable[[], Any | Awaitable[Any]], *, last: bool = False, msg: str = '') -> None:
    global _REGISTERED

    if last is not True and last is not False:
        raise ValueError()

    if not isinstance(msg, str):
        raise TypeError()

    if not msg:
        msg = f'{func.__module__}.{get_obj_name(func)}'

    for existing in _REGISTERED:
        if existing.func is func:
            # If it's the same thing we don't call it multiple times
            if existing.msg == msg and existing.last == last:
                return None

            log.warning(f'Function {func} is already registered with a different message!')
            log.warning(f'  - {existing.msg:s}')
            log.warning(f'  - {msg:s}')
            return None

    if iscoroutinefunction(func):
        _REGISTERED += (ShutdownAwaitable(func=func, last=last, msg=msg), )
    elif isinstance(func, (FunctionType, MethodType, BuiltinMethodType)):
        _REGISTERED += (ShutdownFunction(func=func, last=last, msg=msg), )
    else:
        raise TypeError()


async def _shutdown() -> None:
    global _REQUESTED

    if _REQUESTED:
        return None
    _REQUESTED = True

    log.debug('Requested shutdown')

    objs = (
        *(obj for obj in _REGISTERED if not obj.last),
        *(obj for obj in _REGISTERED if obj.last),
        # shutdown of the event loop has to be the last thing that is done
        # since stopping of the loop exits the program
        ShutdownFunction(func=loop.stop, msg='Stopping asyncio loop', last=True)
    )

    for obj in objs:
        try:
            log.debug(f'{obj.msg}')
            await obj.run()
            log.debug('-> done!')
            await sleep(0.02)
        except Exception as ex:  # noqa: PERF203
            log.error(ex)
            tb = traceback.format_exc().splitlines()
            for line in tb:
                log.error(line)

    log.debug('Shutdown complete')


def request() -> None:
    create_task(_shutdown())


def is_requested() -> bool:
    return _REQUESTED


def register_signal_handler() -> None:
    def shutdown_handler(sig, frame) -> None:
        print('Shutting down ...')
        request()

    # register shutdown helper
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Final, TypeVar

import HABApp
from HABApp.core.connections import BaseConnection
from HABApp.core.connections._definitions import connection_log


if TYPE_CHECKING:
    from collections.abc import Generator


T = TypeVar('T', bound=BaseConnection)


class ConnectionManager:
    def __init__(self) -> None:
        self.connections: dict[str, BaseConnection] = {}

    def add(self, connection: T) -> T:
        if connection.name in self.connections:
            msg = f'Connection {connection.name:s} already exists!'
            raise ValueError(msg)

        self.connections[connection.name] = connection
        connection_log.debug(f'Added {connection.name:s}')
        return connection

    def get(self, name: str) -> BaseConnection:
        return self.connections[name]

    def get_names(self) -> Generator[str, None, None]:
        yield from self.connections.keys()

    def remove(self, name: str) -> None:
        con = self.get(name)
        if not con.is_shutdown:
            raise ValueError()
        self.connections.pop(name)

    async def on_application_shutdown(self) -> None:
        for c in self.connections.values():
            c.on_application_shutdown()

        tasks = [t.advance_status_task.wait() for t in self.connections.values()]
        await asyncio.gather(*tasks)

    def application_startup_complete(self) -> None:
        for c in self.connections.values():
            with HABApp.core.wrapper.ExceptionToHABApp(logger=c.log):
                c.application_startup_complete()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}>'


connection_manager: Final = ConnectionManager()



from eascheduler import add_holiday, get_holiday_name, get_holidays_by_name, get_sun_position, is_holiday, pop_holiday
from eascheduler.builder import FilterBuilder as filter
from eascheduler.builder import TriggerBuilder as trigger
from whenever import hours, milliseconds, minutes, seconds

from HABApp.core.lib import InstantView



import logging
import logging.config
import sys
from pathlib import Path
from queue import Queue
from typing import Any

from easyconfig.yaml import yaml_safe as _yaml_safe

from HABApp.config.config import CONFIG
from HABApp.config.errors import AbsolutePathExpected
from HABApp.config.logging import rotate_file
from HABApp.core.const.const import PYTHON_312, PYTHON_313
from HABApp.core.const.log import TOPIC_EVENTS

from .buffered_logger import BufferedLogger
from .queue_handler import HABAppQueueHandler, SimpleQueue


def fix_old_logger_location(handlers_cfg: dict, log: BufferedLogger) -> None:
    src = 'HABApp.core.lib.handler.MidnightRotatingFileHandler'
    dst = 'HABApp.config.logging.MidnightRotatingFileHandler'

    # fix filenames
    for handler, cfg in handlers_cfg.items():
        # migrate handler
        if cfg.get('class', '-') == src:
            cfg['class'] = dst
            log.warning(f'Replaced class for handler "{handler:s}" with {dst:s}')


def fix_log_filenames(handlers_cfg: dict) -> None:
    for cfg in handlers_cfg.values():
        if (filename := cfg.get('filename')) is None:
            continue

        # fix encoding for FileHandlers - we always log utf-8
        if 'file' in cfg.get('class', '').lower() and cfg.get('encoding', '') != 'utf-8':
            cfg['encoding'] = 'utf-8'

        # make Filenames absolute path in the log folder if not specified
        p = Path(filename)
        if not p.is_absolute():
            # Our log folder ist not yet converted to path -> it is not loaded
            if not CONFIG.directories.logging.is_absolute():
                raise AbsolutePathExpected()

            # Use defined parent folder
            p = (CONFIG.directories.logging / p).resolve()
            cfg['filename'] = str(p)


def remove_memory_handler_from_cfg(handlers_cfg: dict, loggers_cfg: dict, log: BufferedLogger) -> None:
    # find memory handlers
    memory_targets = {}
    for handler, handler_cfg in handlers_cfg.items():
        if handler_cfg.get('class', '') == 'logging.handlers.MemoryHandler':
            log.error(f'"logging.handlers.MemoryHandler" is no longer required. Please remove from config ({handler})!')
            if 'target' in handler_cfg:
                memory_targets[handler] = handler_cfg['target']

    # remove them from config
    for h_name in memory_targets:
        handlers_cfg.pop(h_name)
        log.warning(f'Removed {h_name:s} from handlers')

    # replace handlers in logger with target
    for logger_name, logger_cfg in loggers_cfg.items():
        logger_handlers = logger_cfg.get('handlers', [])
        for i, logger_handler in enumerate(logger_handlers):
            if (replacement_handler := memory_targets.get(logger_handler)) is None:
                continue
            log.warning(f'Replaced {logger_handler} with {replacement_handler} for logger {logger_name}')
            logger_handlers[i] = replacement_handler


def load_logging_file(path: Path) -> dict[str, Any] | None:
    if not path.is_file():
        return None

    with path.open('r', encoding='utf-8') as file:
        cfg: dict[str, Any] = _yaml_safe.load(file)
    return cfg


def rotate_handler_files(handlers_cfg: dict) -> None:
    for cfg in handlers_cfg.values():
        if (filename := cfg.get('filename')) is None:
            continue
        if (backup_count := cfg.get('backupCount')) is None:
            continue
        file = Path(filename)

        # If the file is empty we do not rotate
        if not file.is_file() or file.stat().st_size <= 10:  # noqa: PLR2004
            continue

        rotate_file(file, backup_count)


def inject_queue_handler(handlers_cfg: dict, loggers_cfg: dict, log: BufferedLogger) -> list[HABAppQueueHandler]:
    if not CONFIG.habapp.logging.use_buffer:
        return []

    prefix = 'HABAppQueue_'

    # Check that the prefix is unique
    for handler_name in handlers_cfg:
        if handler_name.startswith(prefix):
            msg = f'Handler may not start with {prefix:s}'
            raise ValueError(msg)

    # replace the event logs with the buffered one
    buffered_handlers = {}
    for log_name, log_cfg in loggers_cfg.items():
        if not log_name.startswith(TOPIC_EVENTS):
            continue
        _handlers = {n: f'{prefix}{n}' for n in log_cfg['handlers']}
        buffered_handlers.update(_handlers)
        log_cfg['handlers'] = list(_handlers.values())

        # ensure propagate is disabled
        if log_cfg.get('propagate', False):
            log.warning(f'Propagate can not be set for {log_name}!')
        log_cfg['propagate'] = False

    if not buffered_handlers:
        return []

    q_handlers: list[HABAppQueueHandler] = []

    for handler_name, buffered_handler_name in buffered_handlers.items():
        # https://github.com/python/cpython/issues/124653
        if PYTHON_313:
            q: SimpleQueue = SimpleQueue() if sys.version_info[:3] > (3, 13, 0) else Queue()
        elif PYTHON_312:
            q = SimpleQueue() if sys.version_info[:3] > (3, 12, 7) else Queue()
        else:
            q: SimpleQueue = SimpleQueue()
        handlers_cfg[buffered_handler_name] = {'class': 'logging.handlers.QueueHandler', 'queue': q}

        qh = HABAppQueueHandler(q, handler_name, f'LogBuffer{handler_name:s}')
        q_handlers.append(qh)

    return q_handlers


def process_custom_levels(cfg: dict[str, Any], log: BufferedLogger) -> None:
    for level, alias in cfg.pop('levels', {}).items():
        if not isinstance(level, int):
            # noinspection PyProtectedMember
            level = logging._nameToLevel[level]  # noqa: PLW2901
        logging.addLevelName(level, str(alias))
        log.debug(f'Added custom log level "{alias!s}" ({level})')


def get_logging_dict(cfg: dict[str, Any] | None,
                     log: BufferedLogger | logging.Logger) -> tuple[dict[str, Any], list[HABAppQueueHandler]]:

    # make file version optional for config file
    if 'version' in cfg:
        log.warning('Entry "version" is no longer required in the logging configuration file')
    else:
        cfg['version'] = 1

    handlers_cfg = cfg.get('handlers', {})
    loggers_cfg = cfg.get('loggers', {})

    fix_old_logger_location(handlers_cfg, log)
    fix_log_filenames(handlers_cfg)
    remove_memory_handler_from_cfg(handlers_cfg, loggers_cfg, log)

    # Rotate files before opening the handlers
    rotate_handler_files(handlers_cfg)

    # Allow the user to set his own logging levels (with aliases)
    process_custom_levels(cfg, log)

    q_handler = inject_queue_handler(handlers_cfg, loggers_cfg, log)
    return cfg, q_handler



from eascheduler.builder.helper import HINT_POS_TIMEDELTA, get_pos_timedelta_secs
from typing_extensions import Self
from whenever import Instant

from HABApp.core.const.hints import TYPE_EVENT_CALLBACK
from HABApp.core.errors import ItemNameNotOfTypeStrError, WrongItemTypeError
from HABApp.core.internals import (
    EventBusListener,
    EventFilterBase,
    get_current_context,
    uses_get_item,
    uses_item_registry,
)
from HABApp.core.internals.item_registry import ItemRegistryItem
from HABApp.core.lib import InstantView

from .base_item_times import ChangedTime, ItemNoChangeWatch, ItemNoUpdateWatch, UpdatedTime
from .tmp_data import add_tmp_data as _add_tmp_data
from .tmp_data import restore_tmp_data as _restore_tmp_data


get_item = uses_get_item()
item_registry = uses_item_registry()


class BaseItem(ItemRegistryItem):
    """BaseItem, all items must inherit from this class
    """

    @classmethod
    def get_item(cls, name: str) -> Self:
        """Returns an already existing item. If it does not exist or has a different item type an exception will occur.

        :param name: Name of the item
        """
        if not isinstance(name, str):
            raise ItemNameNotOfTypeStrError.from_value(name)
        item = get_item(name)

        if not isinstance(item, cls):
            raise WrongItemTypeError.from_item(item, cls)
        return item

    def __init__(self, name: str) -> None:
        super().__init__(name)

        _now = Instant.now()
        self._last_change: ChangedTime = ChangedTime(self._name, _now)
        self._last_update: UpdatedTime = UpdatedTime(self._name, _now)

    @property
    def last_change(self) -> InstantView:
        """
        :return: Timestamp of the last time when the item has been changed (read only)
        """
        return InstantView(self._last_change.instant)

    @property
    def last_update(self) -> InstantView:
        """
        :return: Timestamp of the last time when the item has been updated (read only)
        """
        return InstantView(self._last_update.instant)

    def __repr__(self) -> str:
        ret = ''
        for k in ['name', 'last_change', 'last_update']:
            ret += f'{", " if ret else ""}{k}: {getattr(self, k)}'
        return f'<{self.__class__.__name__} {ret:s}>'

    def watch_change(self, secs: HINT_POS_TIMEDELTA) -> ItemNoChangeWatch:
        """Generate an event if the item does not change for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        secs = round(get_pos_timedelta_secs(secs), 1)
        return self._last_change.add_watch(secs)

    def watch_update(self, secs: HINT_POS_TIMEDELTA) -> ItemNoUpdateWatch:
        """Generate an event if the item does not receive and update for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        secs = round(get_pos_timedelta_secs(secs), 1)
        return self._last_update.add_watch(secs)

    def listen_event(self, callback: TYPE_EVENT_CALLBACK,
                     event_filter: EventFilterBase | None = None) -> EventBusListener:
        """
        Register an event listener which listens to all event that the item receives

        :param callback: callback that accepts one parameter which will contain the event
        :param event_filter: Event filter. This is typically :class:`~HABApp.core.events.ValueUpdateEventFilter` or
            :class:`~HABApp.core.events.ValueChangeEventFilter` which will also trigger on changes/update from openhab
            or mqtt. Additionally, it can be an instance of :class:`~HABApp.core.events.EventFilter` which additionally
            filters on the values of the event. It is also possible to group filters logically with, e.g.
            :class:`~HABApp.core.events.AndFilterGroup` and :class:`~HABApp.core.events.OrFilterGroup`
        """
        return get_current_context().rule.listen_event(self._name, callback=callback, event_filter=event_filter)

    def _on_item_added(self) -> None:
        """This function gets automatically called when the item is added to the item registry
        """
        _restore_tmp_data(self)

    def _on_item_removed(self) -> None:
        """This function gets automatically called when the item is removed from the item registry
        """
        _add_tmp_data(self)



import logging
import re
import typing
from asyncio import sleep
from pathlib import Path

import HABApp
import HABApp.__cmd_args__ as cmd_args
from HABApp.core import shutdown
from HABApp.core.connections import Connections
from HABApp.core.files.errors import AlreadyHandledFileError
from HABApp.core.internals import uses_item_registry
from HABApp.core.internals.proxy import uses_file_manager
from HABApp.core.internals.wrapped_function import wrap_func
from HABApp.core.logger import log_warning
from HABApp.core.wrapper import log_exception
from HABApp.rule_manager.rule_file import RuleFile


log = logging.getLogger('HABApp.Rules')

item_registry = uses_item_registry()
file_manager = uses_file_manager()


class RuleManager:

    def __init__(self, parent) -> None:
        assert isinstance(parent, HABApp.runtime.Runtime)
        self.runtime = parent

        self.files: typing.Dict[str, RuleFile] = {}

    async def setup(self):

        # shutdown
        shutdown.register(self.shutdown, msg='Cancel rule schedulers')

        if cmd_args.DO_BENCH:
            from HABApp.rule_manager.benchmark import BenchFile
            self.files['bench'] = file = BenchFile(self)
            ok = await wrap_func(file.load).async_run()
            if not ok:
                log.error('Failed to load Benchmark!')
                shutdown.request()
                return None
            await file.check_all_rules()
            return None

        path = HABApp.CONFIG.directories.rules
        prefix = 'rules/'

        file_manager.add_handler(
            self.__class__.__name__, log, prefix=prefix,
            on_load=self.request_file_load, on_unload=self.request_file_unload
        )
        file_manager.add_folder(
            prefix, path, priority=0, pattern=re.compile(r'.py$', re.IGNORECASE), name='rules-python'
        )

        # Initial loading of rules
        HABApp.core.internals.wrap_func(self.load_rules_on_startup, logger=log).run()

    async def load_rules_on_startup(self):

        if HABApp.CONFIG.openhab.general.wait_for_openhab:
            c = Connections.get('openhab')
            while not (c.is_shutdown or c.is_disabled or c.is_online):
                await sleep(1)
        else:
            await sleep(1)

        # if we want to shut down we don't load the rules
        if shutdown.is_requested():
            return None

        # trigger event for every file
        await file_manager.get_file_watcher().load_files(dispatcher_name_include=r'^rules.*$')
        return None

    @log_exception
    def get_rule(self, rule_name):
        found = []
        for file in self.files.values():
            if rule_name is None:
                for rule in file.rules.values():
                    found.append(rule)
            else:
                if rule_name in file.rules:
                    found.append(file.rules[rule_name])

        # if we want all return them
        if rule_name is None:
            return found

        # if we want a special one throw error
        if not found:
            raise KeyError(f'No Rule with name "{rule_name}" found!')
        return found if len(found) > 1 else found[0]

    async def request_file_unload(self, name: str, path: Path) -> None:
        path_str = str(path)

        # Only unload already loaded files
        if path_str not in self.files:
            log_warning(log, f'Rule file {path} is not yet loaded and therefore can not be unloaded')
            return None

        log.debug(f'Removing file: {name}')
        rule = self.files.pop(path_str)

        await rule.unload()

    async def request_file_load(self, name: str, path: Path) -> None:
        path_str = str(path)

        # if we want to shut down we don't load the rules
        if shutdown.is_requested():
            log.debug(f'Skip load of {name:s} because of shutdown')
            return None

        # Only load existing files
        if not path.is_file():
            log_warning(log, f'Rule file {name} ({path}) does not exist and can not be loaded!')
            return None

        log.debug(f'Loading file: {name}')
        self.files[path_str] = file = RuleFile(self, name, path)

        ok = await wrap_func(file.load).async_run()
        if not ok:
            self.files.pop(path_str)
            log.warning(f'Failed to load {path_str}!')
            raise AlreadyHandledFileError()

        log.debug(f'File {name} successfully loaded!')

        # Do simple checks which prevent errors
        await file.check_all_rules()

    async def shutdown(self) -> None:
        for f in self.files.values():
            await f.unload()



from __future__ import annotations

import random
import warnings
from collections.abc import Callable, Hashable, Iterable, Mapping
from datetime import datetime as dt_datetime
from typing import TYPE_CHECKING, Any, Final, TypeAlias

from eascheduler.builder import FilterBuilder, JobBuilder, TriggerBuilder
from eascheduler.builder.helper import HINT_INSTANT, HINT_TIMEDELTA, get_instant, get_pos_timedelta_secs
from eascheduler.builder.triggers import TriggerObject, _get_producer
from eascheduler.executor import ExecutorBase
from eascheduler.jobs import CountdownJob, DateTimeJob, OneTimeJob
from eascheduler.schedulers.async_scheduler import AsyncScheduler
from typing_extensions import ParamSpec, Self, override

from HABApp.core.asyncio import create_task_from_async, run_func_from_async
from HABApp.core.const import loop
from HABApp.core.internals import Context, wrap_func
from HABApp.core.internals.wrapped_function.wrapped_async import WrappedAsyncFunction
from HABApp.rule.scheduler.job_ctrl import CountdownJobControl, DateTimeJobControl, OneTimeJobControl


if TYPE_CHECKING:

    from HABApp.core.internals.wrapped_function.wrapped_thread import WrappedThreadFunction
    from HABApp.rule_ctx import HABAppRuleContext


HINT_CB_P = ParamSpec('HINT_CB_P')
HINT_CB: TypeAlias = Callable[HINT_CB_P, Any]


class WrappedSyncExecutor(ExecutorBase):
    def __init__(self, func: WrappedThreadFunction,
                 args: Iterable = (), kwargs: Mapping[str, Any] | None = None) -> None:
        self._func: Final = func
        self._args: Final = args
        self._kwargs: Final = kwargs if kwargs is not None else {}

    @override
    def execute(self) -> None:
        self._func.run(*self._args, **self._kwargs)


class WrappedAsyncExecutor(ExecutorBase):
    def __init__(self, func: WrappedAsyncFunction,
                 args: Iterable = (), kwargs: Mapping[str, Any] | None = None) -> None:
        self._func: Final = func
        self._args: Final = args
        self._kwargs: Final = kwargs if kwargs is not None else {}

    @override
    def execute(self) -> None:
        create_task_from_async(
            self._func.async_run(*self._args, **self._kwargs),
            name=self._func.name
        )


def wrapped_func_executor(func: Any, args: Iterable = (), kwargs: Mapping[str, Any] | None = None) -> ExecutorBase:
    if isinstance(func, WrappedAsyncFunction):
        return WrappedAsyncExecutor(func, args, kwargs)
    return WrappedSyncExecutor(func, args, kwargs)


class AsyncHABAppScheduler(AsyncScheduler):

    @override
    def set_enabled(self, enabled: bool) -> Self:
        return run_func_from_async(super().set_enabled, enabled)


class HABAppJobBuilder:
    def __init__(self, context: HABAppRuleContext) -> None:
        self._habapp_rule_ctx: Context = context
        self._scheduler: Final = AsyncHABAppScheduler(event_loop=loop, enabled=False)

        self._builder: Final = JobBuilder(self._scheduler, wrapped_func_executor)

        self.trigger: Final = TriggerBuilder
        self.filter: Final = FilterBuilder

    def countdown(self, secs: HINT_TIMEDELTA, callback: HINT_CB,
                  *args: HINT_CB_P.args,
                  job_id: Hashable | None = None, **kwargs: HINT_CB_P.kwargs) -> CountdownJobControl:
        """Create a job that count town a certain time and then execute.

        :param secs: countdown time in seconds
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param job_id:
        :param kwargs: |param_scheduled_cb_kwargs|
        :return: Created job
        """
        callback = wrap_func(callback, context=self._habapp_rule_ctx)

        job = CountdownJob(wrapped_func_executor(callback, args, kwargs), get_pos_timedelta_secs(secs), job_id=job_id)
        run_func_from_async(job.link_scheduler, self._scheduler)
        return CountdownJobControl(job)

    def once(self, instant: HINT_INSTANT, callback: HINT_CB,
             *args: HINT_CB_P.args,
             job_id: Hashable | None = None, **kwargs: HINT_CB_P.kwargs) -> OneTimeJobControl:
        """Create a job that runs once.

        :param instant: countdown time in seconds
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param job_id:
        :param kwargs: |param_scheduled_cb_kwargs|
        :return: Created job
        """
        callback = wrap_func(callback, context=self._habapp_rule_ctx)

        job = OneTimeJob(wrapped_func_executor(callback, args, kwargs), get_instant(instant), job_id=job_id)
        run_func_from_async(job.link_scheduler, self._scheduler)
        return OneTimeJobControl(job)

    def at(self, trigger: TriggerObject, callback: HINT_CB,
           *args: HINT_CB_P.args,
           job_id: Hashable | None = None, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        """Create a job that will run when a provided trigger occurs.

        :param trigger:
        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param job_id:
        :param kwargs: |param_scheduled_cb_kwargs|
        :return: Created job
        """

        # at produces not reoccurring executions, try to make migration graceful
        if not isinstance(trigger, TriggerObject):
            warnings.warn(
                'self.run.at must be called with a Trigger. Use self.run.once to schedule a single execution',
                DeprecationWarning, stacklevel=2
            )
            return self.once(trigger, callback, *args, job_id=job_id, **kwargs)

        callback = wrap_func(callback, context=self._habapp_rule_ctx)

        job = DateTimeJob(wrapped_func_executor(callback, args, kwargs), _get_producer(trigger), job_id=job_id)
        run_func_from_async(job.link_scheduler, self._scheduler)
        return DateTimeJobControl(job)

    # ------------------------------------------------------------------------------------------------------------------
    # convenience functions
    # ------------------------------------------------------------------------------------------------------------------
    def soon(self, callback: HINT_CB,
             *args: HINT_CB_P.args,
             job_id: Hashable | None = None, **kwargs: HINT_CB_P.kwargs) -> OneTimeJobControl:
        """
        Run the callback as soon as possible.

        :param callback: |param_scheduled_cb|
        :param args: |param_scheduled_cb_args|
        :param kwargs: |param_scheduled_cb_kwargs|
        """
        return self.once(None, callback, *args, job_id=job_id, **kwargs)

    # ------------------------------------------------------------------------------------------------------------------
    # deprecated functions
    # ------------------------------------------------------------------------------------------------------------------
    def every(self, start_time, interval,
              callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:

        warnings.warn(
            'self.run.every is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )

        return self.at(
            TriggerBuilder.interval(start_time, interval),
            callback, *args, **kwargs
        )

    def on_day_of_week(self, time, weekdays,
                       callback: HINT_CB, *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_day_of_week is deprecated. Use self.run.at in combination with a trigger and a filter',
            DeprecationWarning, stacklevel=2
        )

        if isinstance(time, dt_datetime):
            time = time.time()

        return self.at(
            TriggerBuilder.time(time).only_at(FilterBuilder.weekdays(weekdays)),
            callback, *args, **kwargs
        )

    def on_every_day(self, time, callback: HINT_CB,
                     *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_every_day is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )

        if isinstance(time, dt_datetime):
            time = time.time()

        return self.at(
            TriggerBuilder.time(time),
            callback, *args, **kwargs
        )

    def on_sunrise(self, callback: HINT_CB,
                   *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_sunrise is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.sunrise(), callback, *args, **kwargs)

    def on_sunset(self, callback: HINT_CB,
                  *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_sunset is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.sunset(), callback, *args, **kwargs)

    def on_sun_dawn(self, callback: HINT_CB,
                  *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_sun_dawn is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.dawn(), callback, *args, **kwargs)

    def on_sun_dusk(self, callback: HINT_CB,
                    *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.on_sun_dusk is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.dusk(), callback, *args, **kwargs)

    def every_minute(self, callback: HINT_CB,
                     *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.every_minute is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.interval(random.randint(0, 60 - 1), 60), callback, *args, **kwargs)

    def every_hour(self, callback: HINT_CB,
                   *args: HINT_CB_P.args, **kwargs: HINT_CB_P.kwargs) -> DateTimeJobControl:
        warnings.warn(
            'self.run.every_hour is deprecated. Use self.run.at in combination with a trigger',
            DeprecationWarning, stacklevel=2
        )
        return self.at(TriggerBuilder.interval(random.randint(0, 3600 - 1), 3600), callback, *args, **kwargs)



from __future__ import annotations

from typing import Final, Literal

from geopy.distance import distance
from typing_extensions import Self


class Point:
    __slots__ = ('_latitude', '_longitude', '_elevation')  # noqa: RUF023

    def __init__(self, latitude: float, longitude: float, elevation: float | None = None) -> None:
        if not -90 <= latitude <= 90:
            msg = 'Latitude must be between -90 and 90'
            raise ValueError(msg)

        if not -180 <= longitude <= 180:
            msg = 'Longitude must be between -180 and 180'
            raise ValueError(msg)

        if elevation is not None and not isinstance(elevation, (float, int)):
            msg = 'Elevation must be int or float'
            raise TypeError(msg)

        self._latitude: Final = latitude
        self._longitude: Final = longitude
        self._elevation: Final = elevation

    @property
    def latitude(self) -> float:
        """Latitude of the point"""
        return self._latitude

    @property
    def longitude(self) -> float:
        """Longitude of the point"""
        return self._longitude

    @property
    def elevation(self) -> float | None:
        """Elevation of the point"""
        return self._elevation

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self._latitude}, {self._longitude}, {self._elevation})'

    def distance(self, other: Point | tuple[float, float, float | None] | tuple[float, float], *,
                 unit: Literal['km', 'kilometers', 'meters', 'm', 'miles', 'mi', 'feet', 'fi'] = 'kilometers') -> float:
        """Calculate the distance between two points

        :param other: Other point or tuple with latitude, longitude and elevation
        :param unit: unit in which the distance should be returned
        :return: distance
        """

        if isinstance(other, Point):
            d = distance((self._latitude, self._longitude), (other._latitude, other._longitude))
            return getattr(d, unit)

        if isinstance(other, tuple) and len(other) in (2, 3):
            return self.distance(Point(*other), unit=unit)

        raise ValueError()

    def __eq__(self, other: Self | tuple[float, float, float | None] | tuple[float, float]) -> bool:
        if isinstance(other, self.__class__):
            return (self._latitude == other._latitude and self._longitude == other._longitude and
                    self._elevation == other._elevation)

        if isinstance(other, tuple):
            # compare directly with tuple so we don't raise and exception
            match len(other):
                case 2:
                    lat, long = other
                    return self._latitude == lat and self._longitude == long and self._elevation is None
                case 3:
                    lat, long, elev = other
                    return self._latitude == lat and self._longitude == long and self._elevation == elev

        return NotImplemented

    def __getitem__(self, item: int | str) -> float:
        if isinstance(item, int):
            if item == 0:
                return self._latitude
            if item == 1:
                return self._longitude
            if item == 2:
                return self._elevation
            raise IndexError()

        if isinstance(item, str):
            if item in ('lat', 'latitude'):
                return self._latitude
            if item in ('long', 'longitude'):
                return self._longitude
            if item in ('elev', 'elevation'):
                return self._elevation
            raise KeyError()

        raise TypeError()



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from typing_extensions import override

from HABApp.core.errors import InvalidItemValueError, ItemValueIsNoneError
from HABApp.core.types import HSB, RGB
from HABApp.openhab.definitions.websockets.item_value_types import (
    HSBTypeModel,
)
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.items.commands import OnOffCommand, PercentCommand


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/ColorItem.java
class ColorItem(OpenhabItem, OnOffCommand, PercentCommand):
    """ColorItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar HSB value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('ColorItem', HSBTypeModel, 'Percent', 'OnOff', 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent(
        'ColorItem', HSBTypeModel, 'Percent', 'OnOff', 'IncreaseDecrease', 'Refresh')
    _state_from_oh_str = staticmethod(HSBTypeModel.get_value_from_state)

    @property
    def hsb(self) -> HSB:
        """HSB value"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v

    @property
    def hue(self) -> float:
        """Hue part of the value"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.hue

    @property
    def saturation(self) -> float:
        """Saturation part of the value"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.saturation

    @property
    def brightness(self) -> float:
        """Brightness part of the value"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.brightness

    @override
    def set_value(self, new_value: RGB | HSB | tuple[float, float, float]) -> bool:
        """Set a new color value without creating events on the event bus

        :param new_value: new value of the item
        :return: True if state has changed
        """

        if isinstance(new_value, HSB):
            hsb = new_value
        elif isinstance(new_value, RGB):
            hsb = new_value.to_hsb()
        elif isinstance(new_value, tuple):
            # map tuples to variables e.g. when calling post_value
            # when processing events instead of three values we get the tuple
            hue, saturation, brightness = new_value
            hsb = HSB(hue, saturation, brightness)
        elif new_value is None:
            hsb = None
        else:
            raise InvalidItemValueError.from_item(self, new_value)

        return super().set_value(new_value=hsb)

    def get_rgb(self) -> RGB:
        """Return a rgb equivalent of the color

        :return: rgb tuple
        """
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.to_rgb()

    def is_on(self) -> bool:
        """Return true if item is on"""
        if (value := self.value) is None:
            return False
        return value.brightness > 0

    def is_off(self) -> bool:
        """Return true if item is off"""
        if (value := self.value) is None:
            return False
        return value.brightness <= 0

    def __repr__(self) -> str:
        if self.value is None:
            return '<Color None>'
        return f'<Color hue: {self.hue}°, saturation: {self.saturation}%, brightness: {self.brightness}%>'



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from typing_extensions import override

from HABApp.core.const import MISSING
from HABApp.core.const.const import _MissingType
from HABApp.core.errors import InvalidItemValueError
from HABApp.openhab.definitions.websockets.item_value_types import StringListTypeModel
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.types import StringList


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/CallItem.java
class CallItem(OpenhabItem):
    """CallItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar StringList value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('CallItem', 'StringList', 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('CallItem', 'Refresh')
    _state_from_oh_str: Final = staticmethod(StringListTypeModel.get_value_from_state)

    @override
    def set_value(self, new_value: tuple[str, ...] | list[str] | StringList | None) -> bool:
        if isinstance(new_value, StringList):
            return super().set_value(new_value)

        if new_value is None:
            return super().set_value(None)

        if isinstance(new_value, (tuple, list)):
            return super().set_value(StringList(new_value))

        raise InvalidItemValueError.from_item(self, new_value)

    @override
    def oh_post_update(self,
                       value: tuple[str, ...] | list[str] | StringList | None | _MissingType = MISSING) -> None:

        if isinstance(value, (tuple, list)):
            return super().oh_post_update(StringList(value))

        return super().oh_post_update(value)



from .context import AutoContextBoundObj, Context, ContextBoundObj, ContextProvidingObj, get_current_context
from .proxy import setup_internals, uses_event_bus, uses_get_item, uses_item_registry, uses_post_event


# isort: split

from .event_bus import EventBus
from .event_filter import EventFilterBase
from .item_registry import ItemRegistry, ItemRegistryItem


# isort: split

from .event_bus_listener import ContextBoundEventBusListener, EventBusListener
from .wrapped_function import WrappedFunctionBase, wrap_func



from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict, Field, PlainSerializer
from pydantic_core import to_json

from HABApp.openhab.events import OpenhabEvent


class BaseModel(_BaseModel):
    model_config = ConfigDict(extra='forbid', strict=True, frozen=True, validate_default=True, validate_assignment=True)


class BaseEvent(BaseModel):
    topic: str

    source: str | None = None
    event_id: str | None = Field(None, alias='eventId')

    def to_event(self) -> OpenhabEvent:
        raise NotImplementedError()


SERIALIZE_TO_JSON_STR = PlainSerializer(lambda x: to_json(x).decode())


MSG_CTR: int = 1


def msg_id() -> str:
    global MSG_CTR

    value = MSG_CTR

    MSG_CTR += 1
    if MSG_CTR >= 1_000_000:  # noqa: PLR2004
        MSG_CTR = 1

    return str(value)


class BaseOutEvent(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    topic: str

    source: str = 'HABApp'
    event_id: str  = Field(default_factory=msg_id, alias='eventId')



from typing import Any

from typing_extensions import override

from HABApp.core.internals import AutoContextBoundObj, EventFilterBase, uses_event_bus
from HABApp.core.internals.event_bus import EventBusBaseListener
from HABApp.core.internals.wrapped_function import WrappedFunctionBase


event_bus = uses_event_bus()


class EventBusListener(EventBusBaseListener):
    def __init__(self, topic: str, callback: WrappedFunctionBase, event_filter: EventFilterBase, **kwargs: Any) -> None:
        super().__init__(topic, **kwargs)

        assert isinstance(callback, WrappedFunctionBase)
        self.func: WrappedFunctionBase = callback
        self.filter: EventFilterBase = event_filter

    def notify_listeners(self, event: Any) -> None:
        if self.filter.trigger(event):
            self.func.run(event)

    def describe(self) -> str:
        return f'"{self.topic}" (filter={self.filter.describe()})'

    def cancel(self) -> None:
        """Stop listening on the event bus"""
        event_bus.remove_listener(self)


class ContextBoundEventBusListener(EventBusListener, AutoContextBoundObj):

    @override
    def _ctx_unlink(self):
        event_bus.remove_listener(self)
        return super()._ctx_unlink()

    @override
    def cancel(self) -> None:
        """Stop listening on the event bus"""
        self._ctx_unlink()



from typing import Any, Final


class EventBusBaseListener:
    def __init__(self, topic: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not isinstance(topic, str):
            raise TypeError()
        if not topic:
            raise ValueError()
        self.topic: Final = topic

    def notify_listeners(self, event: Any) -> None:
        raise NotImplementedError()

    def describe(self) -> str:
        raise NotImplementedError()



from __future__ import annotations

from enum import Enum, auto
from hashlib import blake2b
from typing import TYPE_CHECKING, Final

from HABApp.core.files.errors import AlreadyHandledFileError, CircularReferenceError, DependencyDoesNotExistError
from HABApp.core.files.file_properties import FileProperties
from HABApp.core.wrapper import process_exception


if TYPE_CHECKING:
    import logging
    from collections.abc import Iterable
    from pathlib import Path

    from HABApp.core.files.manager import FileManager, FileTypeHandler


class FileState(Enum):
    LOADED = auto()
    FAILED = auto()

    DEPENDENCIES_OK = auto()
    DEPENDENCIES_MISSING = auto()
    DEPENDENCIES_ERROR = auto()

    # if the file should be automatically unloaded
    UNLOAD_PENDING = auto()

    # initial and last states
    PENDING = auto()
    PROPERTIES_INVALID = auto()  # Properties could not be parsed
    REMOVED = auto()

    def __str__(self) -> str:
        return str(self.name)


class HABAppFile:

    @staticmethod
    def create_checksum(text: str) -> bytes:
        b = blake2b()
        b.update(text.encode())
        return b.digest()

    def __init__(self, name: str, path: Path, checksum: bytes, properties: FileProperties | None) -> None:
        self.name: Final = name
        self.path: Final = path
        self.checksum: Final = checksum
        self.properties: Final = properties if properties is not None else FileProperties()
        self._state: FileState = FileState.PENDING if properties is not None else FileState.PROPERTIES_INVALID

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.name} state: {self._state}>'

    def set_state(self, new_state: FileState, manager: FileManager) -> None:
        if self._state is new_state:
            return None

        self._state = new_state
        manager.file_state_changed(self, str(new_state))

    def _check_circ_refs(self, stack: tuple[str, ...], prop: str, manager: FileManager) -> None:
        c: Iterable[str] = getattr(self.properties, prop)
        for f in c:
            _stack = stack + (f, )
            if f in stack:
                raise CircularReferenceError(_stack)

            next_file = manager.get_file(f)
            if next_file is not None:
                next_file._check_circ_refs(_stack, prop, manager)

    def _check_properties(self, manager: FileManager, log: logging.Logger) -> None:
        # check dependencies
        missing = {name for name in self.properties.depends_on if manager.get_file(name) is None}
        if missing:
            one = len(missing) == 1
            msg = (f'File {self.path} depends on file{"" if one else "s"} that '
                   f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(missing))}')
            raise DependencyDoesNotExistError(msg)

        # check reload
        missing = {name for name in self.properties.reloads_on if manager.get_file(name) is None}
        if missing:
            one = len(missing) == 1
            log.warning(f'File {self.path} reloads on file{"" if one else "s"} that '
                        f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(missing))}')

    def check_properties(self, manager: FileManager, log: logging.Logger, *, log_msg: bool = False) -> None:
        if self._state is not FileState.PENDING and self._state is not FileState.DEPENDENCIES_ERROR:
            return None

        try:
            self._check_properties(manager, log)
        except DependencyDoesNotExistError as e:
            if log_msg:
                log.error(e.msg)
            return self.set_state(FileState.DEPENDENCIES_ERROR, manager)

        try:
            # check for circular references
            self._check_circ_refs((self.name, ), 'depends_on', manager)
            self._check_circ_refs((self.name, ), 'reloads_on', manager)
        except CircularReferenceError as e:
            log.error(f'Circular reference: {" -> ".join(e.stack)}')
            return self.set_state(FileState.DEPENDENCIES_ERROR, manager)

        # Check if we can already load it
        new_state = FileState.DEPENDENCIES_OK if not self.properties.depends_on else FileState.DEPENDENCIES_MISSING
        self.set_state(new_state, manager)
        return None

    def check_dependencies(self, manager: FileManager) -> None:
        if self._state is not FileState.DEPENDENCIES_MISSING:
            return None

        for name in self.properties.depends_on:
            if (file := manager.get_file(name)) is None:
                return None
            if file._state is not FileState.LOADED:
                return None

        self.set_state(FileState.DEPENDENCIES_OK, manager)
        return None

    def can_be_loaded(self) -> bool:
        return self._state is FileState.DEPENDENCIES_OK

    def state_unload_pending(self) -> bool:
        return self._state is FileState.UNLOAD_PENDING

    def can_be_removed(self) -> bool:
        return self._state is FileState.REMOVED

    def can_be_unloaded(self) -> bool:
        # If we are loaded we are either FAILED or LOADED.
        # With failed the handler will have already unloaded the file
        return self._state is FileState.LOADED

    async def load(self, handler: FileTypeHandler, manager: FileManager) -> None:
        if not self.can_be_loaded():
            msg = f'File {self.name} can not be loaded because current state is {self._state}!'
            raise ValueError(msg)

        try:
            await handler.on_load(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(handler.on_load, e, logger=handler.logger)
            self.set_state(FileState.FAILED, manager)
            return None

        self.set_state(FileState.LOADED, manager)
        return None

    async def unload(self, handler: FileTypeHandler, manager: FileManager) -> None:
        try:
            await handler.on_unload(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(handler.on_unload, e, logger=handler.logger)

            if self._state is FileState.UNLOAD_PENDING:
                self.set_state(FileState.PENDING, manager)
            else:
                self.set_state(FileState.FAILED, manager)
            return None

        if self._state is FileState.UNLOAD_PENDING:
            self.set_state(FileState.PENDING, manager)
        else:
            self.set_state(FileState.REMOVED, manager)
        return None

    def file_state_changed(self, file: HABAppFile, manager: FileManager) -> None:
        name = file.name
        if name in self.properties.reloads_on:
            if self.can_be_unloaded():
                self.set_state(FileState.UNLOAD_PENDING, manager)
            else:
                self.set_state(FileState.PENDING, manager)



import operator as _operator
from collections.abc import Callable
from typing import Any, Final

from HABApp.core.const import MISSING


CMP_OPS: Final[dict[str, Callable[[Any, Any], bool]]] = {
    'lt': _operator.lt, 'lower_than': _operator.lt,
    'le': _operator.le, 'lower_equal': _operator.le,
    'eq': _operator.eq, 'equal': _operator.eq,
    'ne': _operator.ne, 'not_equal': _operator.ne,
    'gt': _operator.gt, 'greater_than': _operator.gt,
    'ge': _operator.ge, 'greater_equal': _operator.ge,

    'is_': _operator.is_,
    'is_not': _operator.is_not,
}


def compare(value: Any, **kwargs) -> bool:

    for name, cmp_value in kwargs.items():
        if cmp_value is MISSING:
            continue
        if CMP_OPS[name](value, cmp_value):
            return True
    return False



from __future__ import annotations

import inspect
import logging
from asyncio import Future as _Future
from asyncio import Task as _Task
from asyncio import run_coroutine_threadsafe as _run_coroutine_threadsafe
from contextvars import ContextVar as _ContextVar
from threading import get_ident
from typing import TYPE_CHECKING, Final
from typing import Any as _Any
from typing import ParamSpec as _ParamSpec
from typing import TypeVar as _TypeVar

import HABApp
from HABApp.core.const import loop
from HABApp.core.const.installation import PYTHON_INSTALLATION_PATHS
from HABApp.core.const.topics import TOPIC_ERRORS
from HABApp.core.lib.helper import get_obj_name


if TYPE_CHECKING:
    from collections.abc import Awaitable as _Awaitable
    from collections.abc import Callable as _Callable
    from collections.abc import Coroutine as _Coroutine


thread_context: Final = _ContextVar('thread_ctx')
thread_ident: Final = get_ident()


class AsyncContextError(Exception):
    def __init__(self, func: _Callable) -> None:
        super().__init__()
        self.func: _Callable = func

    def __str__(self) -> str:
        return f'Function "{get_obj_name(self.func)}" may not be called from an async context!'


def thread_error_msg() -> None:
    call_stack = []
    for frame in reversed(inspect.stack()):
        for name in PYTHON_INSTALLATION_PATHS:
            if frame.filename.startswith(name):
                break
        else:
            call_stack.append(frame)

    error_msg = [
        'Thread usage detected but no thread marker "@in_thread" was used!',
        'See https://habapp.readthedocs.io/en/latest/troubleshooting.html for more information.',
    ]

    if len(call_stack) > 1:
        error_msg.append('Call stack:')
    if call_stack:
        error_msg.extend(f'{frame.filename:s}:{frame.lineno:d} {frame.function:s}' for frame in call_stack)

    log = logging.getLogger('HABApp')
    for line in error_msg:
        log.error(line)

    # send event so user can receive notification
    HABApp.core.EventBus.post_event(TOPIC_ERRORS, '\n'.join(error_msg))


def _in_thread() -> bool:
    thread_ctx = thread_context.get(None) is not None
    same_ident = get_ident() == thread_ident

    # both markers point to async
    if not thread_ctx and same_ident:
        return False

    # both markers point to thread
    if thread_ctx and not same_ident:
        return True

    # markers are in different states -> log error but assume thread
    thread_error_msg()
    return True


_tasks = set()


_T = _TypeVar('_T')


def create_task(coro: _Coroutine[_Any, _Any, _T], name: str | None = None) -> _Future[_T]:
    # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    if _in_thread():
        f = _run_coroutine_threadsafe(_async_execute_awaitable(coro), loop)
        _tasks.add(f)
        f.add_done_callback(_tasks.discard)
        return f

    t = loop.create_task(coro, name=name)
    _tasks.add(t)
    t.add_done_callback(_tasks.discard)
    return t


def create_task_from_async(coro: _Coroutine[_Any, _Any, _T], name: str | None = None) -> _Task[_T]:
    t = loop.create_task(coro, name=name)
    _tasks.add(t)
    t.add_done_callback(_tasks.discard)
    return t


def run_coro_from_thread(coro: _Coroutine[_Any, _Any, _T], calling: _Callable) -> _T:
    # This function call is blocking, so it can't be called in the async context
    if not _in_thread():
        raise AsyncContextError(calling)

    fut = _run_coroutine_threadsafe(_async_execute_awaitable(coro), loop)
    return fut.result()


_P = _ParamSpec('_P')


def run_func_from_async(func: _Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> _T:
    """Runs a function from an async context"""

    # we already have an async context
    if not _in_thread():
        return func(*args, **kwargs)

    # we are in a thread, that's why we can wait (and block) for the future
    future = _run_coroutine_threadsafe(_async_execute_func(func, *args, **kwargs), loop)
    return future.result()


async def _async_execute_func(func: _Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> _T:
    """Helper coroutine to execute a function from the event loop"""

    ctx = thread_context.set(None)
    try:
        return func(*args, **kwargs)
    finally:
        thread_context.reset(ctx)


async def _async_execute_awaitable(awaitable: _Awaitable[_T]) -> _T:
    """Helper coroutine to execute a coroutine from the event loop and wait for the result"""

    ctx = thread_context.set(None)
    try:
        return await awaitable
    finally:
        thread_context.reset(ctx)



from HABApp.core import asyncio, const, errors, lib, shutdown, types


# isort: split

# The connection manager has no dependencies - that's why we can set it up before the internals
from HABApp.core.connections import Connections


# isort: split
from HABApp.core import internals, logger, wrapper


# isort: split

import HABApp.core.events
import HABApp.core.files
import HABApp.core.items


# isort: split

Items: 'HABApp.core.internals.ItemRegistry' = internals.proxy.ConstProxyObj('ItemRegistry')
EventBus: 'HABApp.core.internals.EventBus' = internals.proxy.ConstProxyObj('EventBus')



from HABApp.openhab.connection.handler.func_async import (
    async_create_item,
    async_get_item,
    async_get_items,
    async_get_root,
    async_get_system_info,
    async_get_thing,
    async_get_things,
    async_get_uuid,
    async_item_exists,
    async_remove_item,
    async_remove_metadata,
    async_set_metadata,
    async_set_thing_cfg,
    async_set_thing_enabled,
)
from HABApp.openhab.connection.plugins import async_post_update, async_send_command



from HABApp.openhab.connection.handler.func_sync import (
    create_item,
    create_link,
    get_item,
    get_link,
    get_persistence_data,
    get_persistence_services,
    get_thing,
    item_exists,
    remove_item,
    remove_link,
    remove_metadata,
    set_metadata,
    set_persistence_data,
    set_thing_enabled,
)
from HABApp.openhab.connection.plugins import post_update, send_command



from __future__ import annotations

from asyncio import Queue, QueueEmpty, sleep
from typing import TYPE_CHECKING, Any, Final, Literal

from HABApp.core.asyncio import run_func_from_async
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import ItemRegistryItem, uses_get_item
from HABApp.core.lib import SingleTask
from HABApp.core.logger import log_error, log_info, log_warning
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.connection.handler import convert_to_oh_str, post, put
from HABApp.openhab.definitions.websockets import ItemCommandSendEvent, ItemStateSendEvent
from HABApp.openhab.definitions.websockets.item_value_types import RawTypeModel


if TYPE_CHECKING:
    from HABApp.openhab.definitions.websockets.base import BaseOutEvent
    from HABApp.openhab.items import OpenhabItem


get_item = uses_get_item()


def empty_queue(queue: Queue) -> None:
    if queue is not None:
        try:
            while True:
                queue.get_nowait()
        except QueueEmpty:
            pass


class OutgoingCommandsPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None) -> None:
        super().__init__(name)

        self.queue: Queue[BaseOutEvent] | None = None
        self.http_queue: Queue[tuple[str, str, bool]] | None = None

        self.task_http_worker: Final = SingleTask(self.http_queue_worker, 'OhHttpQueueWorker')
        self.task_watcher_websocket: Final = SingleTask(self.websocket_queue_watcher, 'OhWebsocketQueueWatcher')
        self.task_watcher_http: Final = SingleTask(self.http_queue_watcher, 'OhHttpQueueWatcher')

    async def on_connected(self, context: OpenhabContext) -> None:
        self.queue: Queue[BaseOutEvent] = context.out_queue
        self.http_queue: Queue[tuple[str, str, bool]] = Queue()

        self.task_http_worker.start()
        self.task_watcher_http.start()
        self.task_http_worker.start()

    async def on_disconnected(self) -> None:
        queue = self.queue
        self.queue = None
        http_queue = self.http_queue
        self.http_queue = None

        await self.task_http_worker.cancel_wait()
        await self.task_watcher_http.cancel_wait()
        await self.task_http_worker.cancel_wait()

        empty_queue(queue)
        empty_queue(http_queue)

    async def http_queue_worker(self) -> None:

        queue: Final = self.http_queue

        while True:
            try:
                while True:
                    item, state, is_cmd = await queue.get()

                    # this check should never be hit
                    if not isinstance(state, str):
                        log_error(
                            self.plugin_connection.log,
                            f'Ignored invalid state for item {item:s}: "{state}" ({type(state)})'
                        )
                        continue

                    if is_cmd:
                        await post(f'/rest/items/{item:s}', data=state)
                    else:
                        await put(f'/rest/items/{item:s}/state', data=state)
            except Exception as e:  # noqa: PERF203
                self.plugin_connection.process_exception(e, 'Outgoing queue worker')

    def async_post_update(self, item: str | ItemRegistryItem, state: Any) -> None:
        if not isinstance(item, str):
            item = item.name
        if not isinstance(state, str):
            state = convert_to_oh_str(state)
        if (queue := self.http_queue) is None:
            return None
        queue.put_nowait((item, state, False))

    def async_send_command(self, item: str | ItemRegistryItem, state: Any) -> None:
        if not isinstance(item, str):
            item = item.name
        if not isinstance(state, str):
            state = convert_to_oh_str(state)
        if (queue := self.http_queue) is None:
            return None
        queue.put_nowait((item, state, True))

    def async_send_websocket_event(self, event: ItemStateSendEvent | ItemCommandSendEvent) -> None:
        if not isinstance(event, (ItemStateSendEvent, ItemCommandSendEvent)):
            msg = f'Invalid event type: {type(event)}'
            raise TypeError(msg)

        # Workaround for big message sizes
        # https://github.com/openhab/openhab-core/issues/4587
        if isinstance(event.payload, RawTypeModel):
            if (http_queue := self.http_queue) is None:
                return None
            # 'openhab/items/<NAME>/<state|command>'
            _, _, name, action = event.topic.split('/')
            http_queue.put_nowait((name, event.payload.value, action == 'command'))
            return None

        if (queue := self.queue) is None:
            return None
        queue.put_nowait(event)

    async def websocket_queue_watcher(self) -> None:
        await self._queue_watcher(self.queue, 'websocket')

    async def http_queue_watcher(self) -> None:
        await self._queue_watcher(self.http_queue, 'http')

    async def _queue_watcher(self, queue: Queue, name: str) -> None:
        log = self.plugin_connection.log
        first_msg_at = 150

        upper = first_msg_at
        lower = -1
        last_info_at = first_msg_at // 2

        while True:
            await sleep(10)
            size = queue.qsize()

            # small log msg
            if size > upper:
                upper = size * 2
                lower = size // 2
                log_warning(log, f'{size} messages in {name:s} queue')
            elif size < lower:
                upper = max(size / 2, first_msg_at)
                lower = size // 2
                if lower <= last_info_at:
                    lower = -1
                    log_info(log, 'queue OK')
                else:
                    log_info(log, f'{size} messages in {name:s} queue')


OUTGOING_PLUGIN: Final = OutgoingCommandsPlugin()
async_post_update: Final = OUTGOING_PLUGIN.async_post_update
async_send_command: Final = OUTGOING_PLUGIN.async_send_command
async_send_websocket_event: Final = OUTGOING_PLUGIN.async_send_websocket_event


def send_websocket_event(event: ItemStateSendEvent | ItemCommandSendEvent) -> None:
    return run_func_from_async(async_send_websocket_event, event)


def try_get_item(item: str | ItemRegistryItem) -> OpenhabItem | None:
    item_name = item.name if isinstance(item, ItemRegistryItem) else item

    if not isinstance(item_name, str):
        msg = f'Invalid item type: {type(item)}'
        raise TypeError(msg)

    try:
        item = get_item(item_name)
    except ItemNotFoundException:
        return None

    return item


def post_update(item: str | ItemRegistryItem, state: Any, *,
                transport: Literal['http', 'websocket'] = 'websocket') -> None:
    """
    Post an update to the item

    :param item: item name or item
    :param state: new item state
    :param transport: transport to use. Websocket is much faster but stricter concerning which types are accepted
    """

    # by default, we use the websocket connection because it's much faster
    if transport != 'websocket' or (item_obj := try_get_item(item)) is None:
        return run_func_from_async(async_post_update, item, state)

    return item_obj.oh_post_update(state)


def send_command(item: str | ItemRegistryItem, command: Any, *,
                transport: Literal['http', 'websocket'] = 'websocket') -> None:
    """
    Send the specified command to the item

    :param item: item name or item
    :param command: command
    :param transport: transport to use. Websocket is much faster but stricter concerning which types are accepted
    """

    # by default, we use the websocket connection because it's much faster
    if transport != 'websocket' or (item_obj := try_get_item(item)) is None:
        return run_func_from_async(async_send_command, item, command)

    return item_obj.oh_send_command(command)



from typing import Any


class EventFilterBase:
    def trigger(self, event: Any) -> bool:
        raise NotImplementedError()

    def describe(self) -> str:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f'<{self.describe()} at 0x{id(self):X}>'



from typing import Annotated, Literal

from pydantic import Field, Json

from HABApp.core.const.const import StrEnum

from .base import SERIALIZE_TO_JSON_STR, BaseEvent, BaseOutEvent
from .item_value_types import OpenHabValueType


class WebsocketTopicEnum(StrEnum):
    REQUEST_FAILED = 'openhab/websocket/response/failed'
    REQUEST_SUCCESS = 'openhab/websocket/response/success'
    HEARTBEAT = 'openhab/websocket/heartbeat'
    TYPE_FILTER = 'openhab/websocket/filter/type'


class WebsocketBaseEvent(BaseEvent):
    type: Literal['WebSocketEvent']


class WebsocketRequestFailedEvent(WebsocketBaseEvent):
    topic: Literal[WebsocketTopicEnum.REQUEST_FAILED]
    payload: str


class WebsocketRequestSuccessEvent(WebsocketBaseEvent):
    topic: Literal[WebsocketTopicEnum.REQUEST_SUCCESS]
    payload: Literal['']


class WebsocketHeartbeatEvent(WebsocketBaseEvent):
    topic: Literal[WebsocketTopicEnum.HEARTBEAT]
    payload: Literal['PONG', 'PING']


class WebsocketTypeFilterEvent(WebsocketBaseEvent):
    topic: Literal[WebsocketTopicEnum.TYPE_FILTER]
    payload: Json[list[str]]


WebsocketEventType = Annotated[
    WebsocketRequestFailedEvent |
    WebsocketRequestSuccessEvent |
    WebsocketHeartbeatEvent |
    WebsocketTypeFilterEvent,
    Field(discriminator='topic')
]


# ----------------------------------------------------------------------------------------------------------------------
# Outgoing events
# ----------------------------------------------------------------------------------------------------------------------
class WebsocketSendTypeFilter(BaseOutEvent):
    type: Literal['WebSocketEvent'] = 'WebSocketEvent'
    topic: Literal[WebsocketTopicEnum.TYPE_FILTER] = WebsocketTopicEnum.TYPE_FILTER
    payload: Annotated[list[str], SERIALIZE_TO_JSON_STR]


class ItemStateSendEvent(BaseOutEvent):
    type: Literal['ItemStateEvent'] = 'ItemStateEvent'
    topic: str
    payload: Annotated[OpenHabValueType, SERIALIZE_TO_JSON_STR]


class ItemCommandSendEvent(BaseOutEvent):
    type: Literal['ItemCommandEvent'] = 'ItemCommandEvent'
    topic: str
    payload: Annotated[OpenHabValueType, SERIALIZE_TO_JSON_STR]



import re
from base64 import b64decode, b64encode
from datetime import datetime
from typing import Annotated, Any, Literal

from fastnumbers import float as fast_float
from fastnumbers import real, try_int, try_real
from pydantic import BaseModel as _BaseModel
from pydantic import ConfigDict, Field, TypeAdapter
from typing_extensions import Self, override
from whenever import Instant, OffsetDateTime, PlainDateTime, SystemDateTime, ZonedDateTime

from HABApp.core.types import HSB, RGB, Point
from HABApp.openhab.types import RawType, StringList
from HABApp.openhab.types.quantity import QuantityFloat, QuantityInt


class BaseModel(_BaseModel):
    model_config = ConfigDict(extra='forbid', strict=True, validate_default=True, validate_assignment=True)


class ItemValueBase(BaseModel):
    def get_value(self) -> Any:
        raise NotImplementedError()

    @staticmethod
    def get_value_from_state(state: str) -> Any:
        raise NotImplementedError()

    @classmethod
    def from_value(cls, value: Any) -> Self | None:
        raise NotImplementedError()


# ----------------------------------------------------------------------------------------------------------------------
# Primitive Types
# ----------------------------------------------------------------------------------------------------------------------

# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/RefreshType.java
class RefreshTypeModel(ItemValueBase):
    type: Literal['Refresh']
    value: Literal['REFRESH']

    @override
    def get_value(self) -> Literal['REFRESH']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: Literal['REFRESH'] | None) -> Self | None:
        if value in ('REFRESH', ):
            return cls(type='Refresh', value='REFRESH')
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/UnDefType.java
class UnDefTypeModel(ItemValueBase):
    type: Literal['UnDef']
    value: Literal['NULL', 'UNDEF']

    @override
    def get_value(self) -> None:
        return None

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: None = None) -> Self | None:
        if value is None:
            return cls(type='UnDef', value='NULL')
        return None


# ----------------------------------------------------------------------------------------------------------------------
# Library Types
# ----------------------------------------------------------------------------------------------------------------------

# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/DateTimeType.java
class DateTimeTypeModel(ItemValueBase):
    type: Literal['DateTime']
    value: str = Field(pattern=r'^\d{4}\-\d\d\-\d\dT\d\d:\d\d:\d\d')

    @override
    def get_value(self) -> datetime:
        value = self.value
        # Currently colon is not supported
        # https://github.com/ariebovenberg/whenever/issues/204
        return OffsetDateTime.parse_common_iso(f'{value[:-2]:s}:{value[-2:]:s}').to_plain().py_datetime()

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> datetime:
        # noinspection PyTypeChecker
        # Is not strict, that's why we can pass a string
        return DateTimeTypeModel(type='DateTime', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: datetime | Instant | PlainDateTime |
                               ZonedDateTime | OffsetDateTime | SystemDateTime) -> Self | None:
        if isinstance(value, datetime):
            return cls(type='DateTime', value=value.isoformat())

        # https://whenever.readthedocs.io/en/latest/overview.html#iso-8601
        if isinstance(value, (Instant, PlainDateTime, ZonedDateTime, OffsetDateTime, SystemDateTime)):
            return cls(type='DateTime', value=value.format_common_iso())

        if isinstance(value, str):
            # try parsing through whenever types and datetime
            for parse in (Instant.parse_common_iso, PlainDateTime.parse_common_iso, ZonedDateTime.parse_common_iso,
                          OffsetDateTime.parse_common_iso, SystemDateTime.parse_common_iso, datetime.fromisoformat):
                try:
                    v = parse(value)
                except ValueError:  # noqa: PERF203
                    pass
                else:
                    return cls.from_value(v)

        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/DecimalType.java
class DecimalTypeModel(ItemValueBase):
    type: Literal['Decimal']
    value: str

    @override
    def get_value(self) -> float:
        return real(self.value)

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> float:
        return DecimalTypeModel(type='Decimal', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: float) -> Self | None:
        if isinstance(value, (int, float)):
            return cls(type='Decimal', value=str(value))
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/HSBType.java
class HSBTypeModel(ItemValueBase):
    type: Literal['HSB']
    value: str = Field(pattern=r'^\d+(\.\d+)?,\d+(\.\d+)?,\d+(\.\d+)?$')

    @override
    def get_value(self) -> HSB:
        h, s, b = self.value.split(',')
        return HSB(real(h), real(s), real(b))

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> HSB:
        return HSBTypeModel(type='HSB', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: HSB | RGB) -> Self | None:
        if isinstance(value, HSB):
            h = value._hue
            s = value._saturation
            b = value._brightness
            out = ','.join((
                str(h) if isinstance(h, int) else f'{h:.2f}',
                str(s) if isinstance(s, int) else f'{s:.2f}',
                str(b) if isinstance(b, int) else f'{b:.2f}',
            ))
            return cls(type='HSB', value=out)

        if isinstance(value, RGB):
            return cls.from_value(value.to_hsb())

        if isinstance(value, (tuple, list)) and len(value) == 3:
            if len(value) != 3 or not all(isinstance(v, (int, float)) for v in value):
                return None
            return cls(
                type='HSB',
                value=','.join(str(v) if isinstance(v, int) else f'{v:.2f}' for v in value)
            )

        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/IncreaseDecreaseType.java
class IncreaseDecreaseTypeModel(ItemValueBase):
    type: Literal['IncreaseDecrease']
    value: Literal['INCREASE', 'DECREASE']

    @override
    def get_value(self) -> Literal['INCREASE', 'DECREASE']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('INCREASE', 'DECREASE'):
            return cls(type='IncreaseDecrease', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/NextPreviousType.java
class NextPreviousTypeModel(ItemValueBase):
    type: Literal['NextPrevious']
    value: Literal['NEXT', 'PREVIOUS']

    @override
    def get_value(self) -> Literal['NEXT', 'PREVIOUS']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('NEXT', 'PREVIOUS'):
            return cls(type='NextPrevious', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/OnOffType.java
class OnOffTypeModel(ItemValueBase):
    type: Literal['OnOff']
    value: Literal['ON', 'OFF']

    @override
    def get_value(self) -> Literal['ON', 'OFF']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> Literal['ON', 'OFF']:
        return OnOffTypeModel(type='OnOff', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('ON', 'OFF'):
            return cls(type='OnOff', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/OnOffType.java
class OpenClosedTypeModel(ItemValueBase):
    type: Literal['OpenClosed']
    value: Literal['OPEN', 'CLOSED']

    @override
    def get_value(self) -> Literal['OPEN', 'CLOSED']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> Literal['OPEN', 'CLOSED']:
        return OpenClosedTypeModel(type='OpenClosed', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('OPEN', 'CLOSED'):
            return cls(type='OpenClosed', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/PercentType.java
class PercentTypeModel(ItemValueBase):
    type: Literal['Percent']
    value: str = Field(pattern=r'^\d{1,3}(\.\d+)?$')

    @override
    def get_value(self) -> float:
        value = real(self.value)
        if 0 <= value <= 100:
            return value

        msg = f'Invalid value "{self.value!r}" for {self.__class__.__name__}'
        raise ValueError(msg)

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> float:
        return PercentTypeModel(type='Percent', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if not isinstance(value, (int, float)):
            return None

        if not 0 <= value <= 100:
            return None

        return cls(type='Percent', value=str(value))


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/PlayPauseType.java
class PlayPauseTypeModel(ItemValueBase):
    type: Literal['PlayPause']
    value: Literal['PLAY', 'PAUSE']

    @override
    def get_value(self) -> Literal['PLAY', 'PAUSE']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('PLAY', 'PAUSE'):
            return cls(type='PlayPause', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/PointType.java
class PointTypeModel(ItemValueBase):
    type: Literal['Point']
    value: str = Field(pattern=r'^-?\d+(\.\d+)?,-?\d+(\.\d+)?(,-?\d+(\.\d+)?)?$')     # two or three entries

    @override
    def get_value(self) -> Point:
        values = self.value.split(',')
        match len(values):
            case 3:
                a, b, c = values
                return Point(real(a), real(b), real(c))
            case 2:
                a, b = values
                return Point(real(a), real(b), None)

        msg = f'Invalid value "{self.value!r}" for {self.__class__.__name__}'
        raise ValueError(msg)

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> Point:
        return PointTypeModel(type='Point', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: Point) -> Self | None:
        if isinstance(value, Point):
            a = value._latitude
            b = value._longitude
            c = value._elevation
            if c is not None:
                return cls(type='Point', value=f'{a},{b},{c}')
            return cls(type='Point', value=f'{a},{b}')

        if isinstance(value, (tuple, list)):
            if len(value) not in (2, 3) or not all(isinstance(v, (int, float)) for v in value):
                return None
            return cls(type='Point', value=','.join(str(v) for v in value))

        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/QuantityType.java
class QuantityTypeModel(ItemValueBase):
    type: Literal['Quantity']
    value: str = Field(pattern=r'^-?\d+(\.\d+)?([eE][+-]?\d+)?( [^ ]+)?$')

    @override
    def get_value(self) -> QuantityInt | QuantityFloat:
        # Number:Dimensionless has no unit in the state
        if ' ' not in (value := self.value):
            unit = ''
        else:
            value, unit = value.split(' ', 1)

        # always try int
        if (nr := try_int(value, on_fail=None)) is not None:
            return QuantityInt(nr, unit)

        nr = fast_float(value)
        return QuantityFloat(nr, unit)

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> float:
        return QuantityTypeModel(type='Quantity', value=state).get_value()

    @classmethod
    def from_value(cls, value: Any) -> Self | None:
        if not isinstance(value, str):
            return None
        if (values := value.split(' ')) and len(values) != 2:
            return None
        if try_real(values[0], on_fail=None) is None:
            return None
        return cls(type='Quantity', value=value)


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/RawType.java
class RawTypeModel(ItemValueBase):
    type: Literal['Raw']
    value: str = Field(pattern=r'^data:image/[a-z_-]+;base64,[A-Za-z0-9+/]*={0,2}$')

    @override
    def get_value(self) -> RawType:
        value = self.value

        # The data is in this format
        # data:image/png;base64,iVBORw0KGgo...
        bin_type, data_str = value.split(';', 1)
        enc_str, bin_str = data_str.split(',', 1)

        if enc_str != 'base64':
            msg = f'Invalid encoding {enc_str!r} for {RawType.__name__:s}'
            raise ValueError(msg)

        return RawType.create(bin_type.removeprefix('data:'), b64decode(bin_str))

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> RawType:
        return RawTypeModel(type='Raw', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: RawType | bytes) -> Self | None:
        if isinstance(value, RawType):
            bin_type = value._type
            if not bin_type.startswith('image/'):
                bin_type = f'image/{bin_type}'
            return cls(type='Raw', value=f'data:{bin_type};base64,{b64encode(value._data).decode()}')

        if isinstance(value, bytes):
            return cls.from_value(RawType.create(None, value))

        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/RewindFastforwardType.java
class RewindFastforwardTypeModel(ItemValueBase):
    type: Literal['RewindFastforward']
    value: Literal['REWIND', 'FASTFORWARD']

    @override
    def get_value(self) -> Literal['REWIND', 'FASTFORWARD']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('REWIND', 'FASTFORWARD'):
            return cls(type='RewindFastforward', value=value)
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/StopMoveType.java
class StopMoveTypeModel(ItemValueBase):
    type: Literal['StopMove']
    value: Literal['STOP', 'MOVE']

    @override
    def get_value(self) -> Literal['STOP', 'MOVE']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('STOP', 'MOVE'):
            return cls(type='StopMove', value=value)
        return None


RE_SPLIT_STRINGLIST = re.compile(r'(?<!\\),')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/StringListType.java
class StringListTypeModel(ItemValueBase):
    type: Literal['StringList']
    value: str

    @override
    def get_value(self) -> tuple[str, ...]:
        if '\\' not in (value := self.value):
            return StringList(value.split(','))
        return StringList(tuple(e.replace('\\,', ',') for e in RE_SPLIT_STRINGLIST.split(value)))

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> tuple[str, ...]:
        return StringListTypeModel(type='StringList', value=state).get_value()

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: tuple[str, ...] | list[str]) -> Self | None:
        if isinstance(value, StringList):
            return cls(type='StringList', value=','.join(v.replace(',', '\\,') for v in value))

        if isinstance(value, (tuple, list)):
            values = []
            for v in value:
                if not isinstance(v, str):
                    return None
                values.append(v.replace(',', '\\,'))
            return cls(type='StringList', value=','.join(values))

        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/StringType.java
class StringTypeModel(ItemValueBase):
    type: Literal['String']
    value: str

    @override
    def get_value(self) -> str:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @staticmethod
    def get_value_from_state(state: str) -> str:
        if not isinstance(state, str):
            raise TypeError()
        return state

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str | BaseModel) -> Self | None:
        if isinstance(value, str):
            return cls(type='String', value=value)
        if isinstance(value, BaseModel):
            return cls(type='String', value=value.model_dump_json(by_alias=True))
        return None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/types/UpDownType.java
class UpDownTypeModel(ItemValueBase):
    type: Literal['UpDown']
    value: Literal['UP', 'DOWN']

    @override
    def get_value(self) -> Literal['UP', 'DOWN']:
        return self.value

    # noinspection PyNestedDecorators
    @override
    @classmethod
    def from_value(cls, value: str) -> Self | None:
        if value in ('UP', 'DOWN'):
            return cls(type='UpDown', value=value)
        return None


OpenHabValueType = Annotated[
    RefreshTypeModel |
    UnDefTypeModel |
    DateTimeTypeModel |
    DecimalTypeModel |
    HSBTypeModel |
    IncreaseDecreaseTypeModel |
    NextPreviousTypeModel |
    OnOffTypeModel |
    OpenClosedTypeModel |
    PercentTypeModel |
    PlayPauseTypeModel |
    PointTypeModel |
    QuantityTypeModel |
    RawTypeModel |
    RewindFastforwardTypeModel |
    StopMoveTypeModel |
    StringListTypeModel |
    StringTypeModel |
    UpDownTypeModel,
    Field(discriminator='type')
]


OpenHabValueTypeAdapter = TypeAdapter[OpenHabValueType](OpenHabValueType)



from __future__ import annotations

from typing import Annotated, Literal

from pydantic import Field, Json
from typing_extensions import Self, override

from HABApp.openhab.events.item_events import GroupStateChangedEvent as TargetGroupStateChangedEvent
from HABApp.openhab.events.item_events import GroupStateUpdatedEvent as TargetGroupStateUpdatedEvent
from HABApp.openhab.events.item_events import ItemAddedEvent as TargetItemAddedEvent
from HABApp.openhab.events.item_events import ItemCommandEvent as TargetItemCommandEvent
from HABApp.openhab.events.item_events import ItemRemovedEvent as TargetItemRemovedEvent
from HABApp.openhab.events.item_events import ItemStateChangedEvent as TargetItemStateChangedEvent
from HABApp.openhab.events.item_events import ItemStateEvent as TargetItemStateEvent
from HABApp.openhab.events.item_events import ItemStatePredictedEvent as TargetItemStatePredictedEvent
from HABApp.openhab.events.item_events import ItemStateUpdatedEvent as TargetItemStateUpdatedEvent
from HABApp.openhab.events.item_events import ItemUpdatedEvent as TargetItemUpdatedEvent

from .base import SERIALIZE_TO_JSON_STR, BaseEvent, BaseModel, BaseOutEvent
from .item_value_types import OpenHabValueType, OpenHabValueTypeAdapter


class ValueChangedPayload(BaseModel):
    type: str
    value: str
    old_type: str = Field(alias='oldType')
    old_value: str = Field(alias='oldValue')


class ItemStateEvent(BaseEvent):
    type: Literal['ItemStateEvent']
    payload: Json[OpenHabValueType]

    @override
    def to_event(self) -> TargetItemStateEvent:
        payload = self.payload
        return TargetItemStateEvent(
            name=self.topic[14:-6], value=payload.get_value()
        )


class ItemStateUpdatedEvent(BaseEvent):
    type: Literal['ItemStateUpdatedEvent']
    payload: Json[OpenHabValueType]

    @override
    def to_event(self) -> TargetItemStateUpdatedEvent:
        payload = self.payload
        return TargetItemStateUpdatedEvent(
            name=self.topic[14:-13], value=payload.get_value()
        )


class ItemStateChangedEvent(BaseEvent):
    type: Literal['ItemStateChangedEvent']
    payload: Json[ValueChangedPayload]

    @override
    def to_event(self) -> TargetItemStateChangedEvent:
        payload = self.payload
        new = (ta := OpenHabValueTypeAdapter).validate_python({'type': payload.type, 'value': payload.value})
        old = ta.validate_python({'type': payload.old_type, 'value': payload.old_value})
        return TargetItemStateChangedEvent(
            name=self.topic[14:-13],
            value=new.get_value(),
            old_value=old.get_value()
        )


class ItemCommandEvent(BaseEvent):
    type: Literal['ItemCommandEvent']
    payload: Json[OpenHabValueType]

    @override
    def to_event(self) -> TargetItemCommandEvent:
        payload = self.payload
        return TargetItemCommandEvent(
            name=self.topic[14:-8], value=payload.get_value()
        )


class PredictedPayloadModel(BaseModel):
    type: str = Field(alias='predictedType')
    value: str = Field(alias='predictedValue')
    is_confirmation: bool = Field(alias='isConfirmation')


class ItemStatePredictedEvent(BaseEvent):
    type: Literal['ItemStatePredictedEvent']
    payload: Json[PredictedPayloadModel]

    @override
    def to_event(self) -> TargetItemStatePredictedEvent:
        payload = self.payload
        value = OpenHabValueTypeAdapter.validate_python({'type': payload.type, 'value': payload.value})
        return TargetItemStatePredictedEvent(
            name=self.topic[14:-15], value=value.get_value(), is_confirmation=payload.is_confirmation
        )


class GroupStateUpdatedEvent(BaseEvent):
    type: Literal['GroupStateUpdatedEvent']
    payload: Json[OpenHabValueType]

    @override
    def to_event(self) -> TargetGroupStateUpdatedEvent:
        parts = self.topic.split('/')
        payload = self.payload
        return TargetGroupStateUpdatedEvent(
            name=parts[2], item=parts[3], value=payload.get_value()
        )


class GroupStateChangedEvent(BaseEvent):
    type: Literal['GroupItemStateChangedEvent']
    payload: Json[ValueChangedPayload]

    @override
    def to_event(self) -> TargetGroupStateChangedEvent:
        parts = self.topic.split('/')
        payload = self.payload
        new = (ta := OpenHabValueTypeAdapter).validate_python({'type': payload.type, 'value': payload.value})
        old = ta.validate_python({'type': payload.old_type, 'value': payload.old_value})
        return TargetGroupStateChangedEvent(
            name=parts[2], item=parts[3],
            value=new.get_value(),
            old_value=old.get_value()
        )


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/ItemDTO.java
class ItemDtoModel(BaseModel):
    type: str
    name: str
    label: str | None = None
    category: str | None = None
    tags: frozenset[str]
    groups: frozenset[str] = Field(alias='groupNames')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/GroupFunctionDTO.java
class GroupFunctionDTOModel(BaseModel):
    name: str
    params: tuple[str, ...]


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/GroupItemDTO.java
class GroupItemDtoModel(ItemDtoModel):
    group_type: str | None = Field(default=None, alias='groupType')
    group_function: GroupFunctionDTOModel | None = Field(default=None, alias='function')


# https://github.com/openhab/openhab-core/blob/ce374252fa2c821103da888302f930c1c127f73c/bundles/org.openhab.core/src/main/java/org/openhab/core/items/events/ItemAddedEvent.java
class ItemAddedEvent(BaseEvent):
    type: Literal['ItemAddedEvent']
    payload: Json[GroupItemDtoModel]

    @override
    def to_event(self) -> TargetItemAddedEvent:
        payload = self.payload
        return TargetItemAddedEvent(
            type=payload.type, name=payload.name, label=payload.label, tags=payload.tags, group_names=payload.groups
        )


class ItemUpdatedEvent(BaseEvent):
    type: Literal['ItemUpdatedEvent']
    payload: Json[tuple[GroupItemDtoModel, GroupItemDtoModel]]

    @override
    def to_event(self) -> TargetItemUpdatedEvent:
        new, old = self.payload
        return TargetItemUpdatedEvent(
            type=new.type, name=new.name, label=new.label, tags=new.tags, group_names=new.groups
        )


class ItemRemovedEvent(BaseEvent):
    type: Literal['ItemRemovedEvent']
    payload: Json[GroupItemDtoModel]

    @override
    def to_event(self) -> TargetItemRemovedEvent:
        payload = self.payload
        return TargetItemRemovedEvent(
            type=payload.type, name=payload.name, label=payload.label, tags=payload.tags, groups=payload.groups
        )


# ----------------------------------------------------------------------------------------------------------------------
# Outgoing events
# ----------------------------------------------------------------------------------------------------------------------
class ItemStateSendEvent(BaseOutEvent):
    type: Literal['ItemStateEvent'] = 'ItemStateEvent'
    topic: str
    payload: Annotated[OpenHabValueType, SERIALIZE_TO_JSON_STR]

    @classmethod
    def create(cls, name: str, payload: OpenHabValueType) -> Self:
        return cls(
            type='ItemStateEvent', topic=f'openhab/items/{name:s}/state', payload=payload
        )


class ItemCommandSendEvent(BaseOutEvent):
    type: Literal['ItemCommandEvent'] = 'ItemCommandEvent'
    topic: str
    payload: Annotated[OpenHabValueType, SERIALIZE_TO_JSON_STR]

    @classmethod
    def create(cls, name: str, payload: OpenHabValueType) -> Self:
        return cls(
            type='ItemCommandEvent', topic=f'openhab/items/{name:s}/command', payload=payload
        )



import logging
import re
from pathlib import Path
from typing import Final

import HABApp
from HABApp.core.files.file import HABAppFile
from HABApp.core.internals.proxy import uses_file_manager

from .parameters import get_parameter_file, remove_parameter_file, set_parameter_file


log = logging.getLogger('HABApp.RuleParameters')

PARAMS_PREFIX: Final = 'params/'
PARAMS_SUFFIX: Final = '.yml'

file_manager = uses_file_manager()


def get_user_name(name: str) -> str:
    return name.removeprefix(PARAMS_PREFIX).removesuffix(PARAMS_SUFFIX)


async def load_file(name: str, path: Path) -> None:
    with path.open(mode='r', encoding='utf-8') as file:
        data = HABApp.core.const.yml.load(file)
    if data is None:
        data = {}

    user_name = get_user_name(name)
    set_parameter_file(user_name, data)

    log.debug(f'Successfully loaded {name}!')


async def unload_file(name: str, path: Path) -> None:
    user_name = get_user_name(name)
    remove_parameter_file(user_name)
    log.debug(f'Removed {user_name}!')


def save_file(file: str):
    assert isinstance(file, str), type(file)
    path = HABApp.CONFIG.directories.params
    if path is None:
        msg = 'Parameter files are disabled! Configure a folder to use them!'
        raise ValueError(msg)

    filename = path / (file + '.yml')

    log.info(f'Updated {filename}')
    with filename.open('w', encoding='utf-8') as outfile:
        HABApp.core.const.yml.dump(get_parameter_file(file), outfile)


class HABAppParameterFile(HABAppFile):
    LOGGER = log
    LOAD_FUNC = load_file
    UNLOAD_FUNC = unload_file


async def setup_param_files() -> bool:
    path = HABApp.CONFIG.directories.params
    if path is None:
        return False

    regex = re.escape(PARAMS_SUFFIX) + '$'

    file_manager.add_handler('ParamFiles', log, prefix=PARAMS_PREFIX, on_load=load_file, on_unload=unload_file)
    file_manager.add_folder(
        PARAMS_PREFIX, path, priority=100, pattern=re.compile(regex, re.IGNORECASE), name='rules-parameters'
    )

    return True



import logging
import threading
from typing import Any

from HABApp.core.const.log import TOPIC_EVENTS

from .base_listener import EventBusBaseListener


event_log = logging.getLogger(TOPIC_EVENTS)
habapp_log = logging.getLogger('HABApp')


class EventBus:
    __slots__ = ('_listeners', '_lock')

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self._listeners: dict[str, tuple[EventBusBaseListener, ...]] = {}

    def post_event(self, topic: str, event: Any) -> None:
        if not isinstance(topic, str):
            msg = f'Topic must be a string! Got {type(topic)}'
            raise TypeError(msg)

        if not isinstance(event, str):
            event_prv = str(event)
        else:
            event_prv = event[:120] + ' ...' if len(event) > 120 else event
            event_prv = "'" + event_prv.replace('\n', '\\n') + "'"

        event_log.info(f'{topic:>20s}: {event_prv}')

        # Notify all listeners
        if (listeners := self._listeners.get(topic)) is not None:
            for listener in listeners:
                listener.notify_listeners(event)
        return None

    def add_listener(self, listener: EventBusBaseListener) -> None:
        if not isinstance(listener, EventBusBaseListener):
            raise TypeError()
        if not isinstance(topic := listener.topic, str):
            raise TypeError()
        if not topic:
            raise ValueError()

        with self._lock:
            item_listeners = self._listeners.get(topic, ())

            # don't add the same listener twice
            if listener in item_listeners:
                habapp_log.warning(f'Event listener for {listener.describe()} has already been added!')
                return None

            # add listener
            self._listeners[topic] = item_listeners + (listener,)
            habapp_log.debug(f'Added event listener for {listener.describe()}')
            return None

    def remove_listener(self, listener: EventBusBaseListener) -> None:
        if not isinstance(listener, EventBusBaseListener):
            raise TypeError()
        if not isinstance(topic := listener.topic, str):
            raise TypeError()
        if not topic:
            raise ValueError()

        with self._lock:
            item_listeners = self._listeners.get(topic, ())

            # print warning if we try to remove it twice
            if listener not in item_listeners:
                habapp_log.warning(f'Event listener for {listener.describe()} has already been removed!')
                return None

            # remove listener
            self._listeners[topic] = tuple(o for o in item_listeners if o is not listener)
            habapp_log.debug(f'Removed event listener for {listener.describe()}')
            return None

    def remove_all_listeners(self) -> None:
        with self._lock:
            self._listeners.clear()



from HABApp.util import functions, multimode
from HABApp.util.cache import ExpiringCache
from HABApp.util.fade import Fade
from HABApp.util.listener_groups import EventListenerGroup
from HABApp.util.rate_limiter import RateLimiter
from HABApp.util.ring_counter import RingCounter, RingCounterTracker
from HABApp.util.statistics import Statistics
from HABApp.util.threshold import Threshold



from typing import Annotated

from pydantic import Field, TypeAdapter

from .channel_events import ChannelDescriptionChangedEvent, ChannelTriggeredEvent
from .item_events import (
    GroupStateChangedEvent,
    GroupStateUpdatedEvent,
    ItemAddedEvent,
    ItemCommandEvent,
    ItemRemovedEvent,
    ItemStateChangedEvent,
    ItemStateEvent,
    ItemStatePredictedEvent,
    ItemStateUpdatedEvent,
    ItemUpdatedEvent,
)
from .thing_events import (
    ThingAddedEvent,
    ThingConfigStatusInfoEvent,
    ThingFirmwareStatusInfoEvent,
    ThingRemovedEvent,
    ThingStatusInfoChangedEvent,
    ThingStatusInfoEvent,
    ThingUpdatedEvent,
)
from .websocket_events import WebsocketEventType


OPENHAB_EVENT_TYPE = Annotated[
    ChannelTriggeredEvent |
    ChannelDescriptionChangedEvent |
    GroupStateChangedEvent |
    GroupStateUpdatedEvent |
    ItemAddedEvent |
    ItemCommandEvent |
    ItemRemovedEvent |
    ItemStateChangedEvent |
    ItemStateEvent |
    ItemStatePredictedEvent |
    ItemStateUpdatedEvent |
    ItemUpdatedEvent |
    ThingAddedEvent |
    ThingConfigStatusInfoEvent |
    ThingFirmwareStatusInfoEvent |
    ThingRemovedEvent |
    ThingStatusInfoChangedEvent |
    ThingStatusInfoEvent |
    ThingUpdatedEvent |
    WebsocketEventType,
    Field(discriminator='type')
]


OPENHAB_EVENT_TYPE_ADAPTER = TypeAdapter[OPENHAB_EVENT_TYPE](OPENHAB_EVENT_TYPE)



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, Final, Generic, TypeVar

from HABApp.core.lib import SingleTask


if TYPE_CHECKING:
    from .base_connection import BaseConnection
    from .plugin_callback import PluginCallbackHandler

T = TypeVar('T', bound='BaseConnection')


class BaseConnectionPlugin(Generic[T]):
    def __init__(self, name: str | None = None) -> None:
        super().__init__()

        if name is None:
            name = self.__class__.__name__
            if name[-6:].lower() == 'plugin':
                name = name[:-6]

        self.plugin_connection: T = None
        self.plugin_name: Final = name
        self.plugin_callbacks: dict[str, PluginCallbackHandler] = {}

    def on_application_shutdown(self) -> None:
        pass


class BaseConnectionPluginConnectedTask(BaseConnectionPlugin[T]):
    def __init__(self, task_coro: Callable[[], Awaitable[Any]],
                 task_name: str, name: str | None = None) -> None:
        super().__init__(name)
        self.task: Final = SingleTask(task_coro, name=task_name)

    async def on_connected(self) -> None:
        self.task.start()

    async def on_disconnected(self) -> None:
        await self.task.cancel_wait()
